from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from typing import Dict


class K8sBlockExecutor(BlockExecutor):
    def execute(
        self,
        analyze_outputs: bool = False,
        global_vars: Dict = None,
        update_status: bool = False,
        **kwargs,
    ) -> None:
        """
        TODO: Implement this method
        """
        self.block.execute_sync(
            analyze_outputs=analyze_outputs,
            global_vars=global_vars,
            update_status=update_status,
        )



from mage_ai.data_cleaner.analysis.calculator import AnalysisCalculator
from mage_ai.data_cleaner.column_types import column_type_detector
from mage_ai.data_cleaner.pipelines.base import DEFAULT_RULES, BasePipeline
from mage_ai.data_cleaner.shared.utils import clean_dataframe
from mage_ai.data_cleaner.statistics.calculator import StatisticsCalculator
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.logger import timer, VerboseFunctionExec


def analyze(df):
    cleaner = DataCleaner()
    return cleaner.analyze(df)


def clean(
    df,
    column_types={},
    df_original=None,
    transform=True,
    rules=DEFAULT_RULES,
    rule_configs={},
    verbose=True,
):
    cleaner = DataCleaner(verbose=verbose)
    return cleaner.clean(
        df,
        column_types=column_types,
        df_original=df_original,
        rules=rules,
        rule_configs=rule_configs,
        transform=transform,
    )


class DataCleaner:
    def __init__(self, verbose=False):
        self.verbose = verbose

    def analyze(self, df, column_types={}, df_original=None):
        """Analyze a dataframe
        1. Detect column types
        2. Calculate statisitics
        3. Calculate analysis
        """
        with timer('data_cleaner.infer_column_types'):
            with VerboseFunctionExec('Inferring variable type from dataset', verbose=self.verbose):
                column_types = column_type_detector.infer_column_types(
                    df, column_types=column_types
                )
        with timer('data_cleaner.clean_series'):
            with VerboseFunctionExec('Converting entries to correct datatype', verbose=self.verbose):
                df = clean_dataframe(df, column_types, dropna=False)
        with timer('data_cleaner.calculate_statistics'):
            statistics = StatisticsCalculator(column_types, verbose=self.verbose).process(
                df, df_original=df_original, is_clean=True
            )
        with timer('data_cleaner.calculate_insights'):
            analysis = AnalysisCalculator(
                df, column_types, statistics, verbose=self.verbose
            ).process(df, is_clean=True)
        return dict(
            cleaned_df=df,
            insights=analysis,
            column_types=column_types,
            statistics=statistics,
        )

    def clean(
        self,
        df,
        column_types={},
        df_original=None,
        transform=True,
        rules=DEFAULT_RULES,
        rule_configs={},
    ):
        df_stats = self.analyze(df, column_types=column_types, df_original=df_original)
        df = df_stats['cleaned_df']
        pipeline = BasePipeline(rules=rules, verbose=self.verbose)
        if df_stats['statistics']['is_timeseries']:
            df = df.sort_values(by=df_stats['statistics']['timeseries_index'], axis=0)
        # TODO: Pass in both cleaned and uncleaned versions of dataset
        with timer('data_cleaner.create_suggested_actions'):
            suggested_actions = pipeline.create_actions(
                df,
                df_stats['column_types'],
                df_stats['statistics'],
                rule_configs=rule_configs,
            )
        with timer('data_cleaner.create_preview_results'):
            pipeline.create_preview_results(df, suggested_actions)
        if transform:
            with timer('data_cleaner.transform_data'):
                df_transformed = pipeline.transform(df, auto=True)
        else:
            df_transformed = df
        return merge_dict(
            df_stats,
            dict(
                df=df_transformed,
                suggestions=suggested_actions,
                pipeline=pipeline,
            ),
        )



from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.io.base import DataSource
from mage_ai.data_preparation.models.block import BlockType
from mage_ai.data_preparation.templates.template import (
    build_template_from_suggestion,
    fetch_template_source,
)
from mage_ai.tests.base_test import TestCase


class TemplateTest(TestCase):
    def test_template_creation(self):
        suggestion = dict(
            title='Remove rows with missing entries',
            message='Delete 3 rows to remove all missing values from the dataset.',
            action_payload=dict(
                action_type='filter',
                action_arguments=['state', 'location'],
                action_options={},
                action_variables=dict(
                    state=dict(feature=dict(column_type='category', uuid='state'), type='feature'),
                    location=dict(
                        feature=dict(column_type='zip_code', uuid='location'), type='feature'
                    ),
                ),
                action_code='state != null and location != null',
                axis='row',
                outputs=[],
            ),
            status='not_applied',
        )
        expected_string = """from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_cleaner.transformer_actions.utils import build_transformer_action
from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def remove_rows_with_missing_entries(df: DataFrame, *args, **kwargs) -> DataFrame:
    \"\"\"
    Transformer Action: Delete 3 rows to remove all missing values from the dataset.
    \"\"\"
    action = {
        "action_type": "filter",
        "action_arguments": [
            "state",
            "location"
        ],
        "action_options": {},
        "action_variables": {
            "state": {
                "feature": {
                    "column_type": "category",
                    "uuid": "state"
                },
                "type": "feature"
            },
            "location": {
                "feature": {
                    "column_type": "zip_code",
                    "uuid": "location"
                },
                "type": "feature"
            }
        },
        "action_code": "state != null and location != null",
        "axis": "row",
        "outputs": []
    }
    return BaseAction(action).execute(df)


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""
        new_string = build_template_from_suggestion(suggestion)
        self.assertEqual(expected_string, new_string)

    def test_template_generation_data_loader_default(self):
        expected_template = """from pandas import DataFrame

if 'data_loader' not in globals():
    from mage_ai.data_preparation.decorators import data_loader
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@data_loader
def load_data(**kwargs) -> DataFrame:
    \"\"\"
    Template code for loading data from any source.

    Returns:
        DataFrame: Returned pandas data frame.
    \"\"\"
    # Specify your data loading logic here

    return DataFrame({})


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config1 = {'data_source': 'default'}
        config2 = {}
        new_template1 = fetch_template_source(BlockType.DATA_LOADER, config1)
        new_template2 = fetch_template_source(BlockType.DATA_LOADER, config2)
        self.assertEqual(expected_template, new_template1)
        self.assertEqual(expected_template, new_template2)

    def test_template_generation_data_loader_specific(self):
        redshift_template = """from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.redshift import Redshift
from os import path
from pandas import DataFrame

if 'data_loader' not in globals():
    from mage_ai.data_preparation.decorators import data_loader
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@data_loader
def load_data_from_redshift(**kwargs) -> DataFrame:
    \"\"\"
    Template for loading data from a Redshift cluster.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/blocks/data_loading.md#redshift
    \"\"\"
    query = 'your_redshift_selection_query'
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    with Redshift.with_config(ConfigFileLoader(config_path, config_profile)) as loader:
        return loader.load(query)


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""
        s3_template = """from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.s3 import S3
from os import path
from pandas import DataFrame

if 'data_loader' not in globals():
    from mage_ai.data_preparation.decorators import data_loader
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@data_loader
def load_from_s3_bucket(**kwargs) -> DataFrame:
    \"\"\"
    Template for loading data from a S3 bucket.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/blocks/data_loading.md#s3
    \"\"\"
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    bucket_name = 'your_bucket_name'
    object_key = 'your_object_key'

    return S3.with_config(ConfigFileLoader(config_path, config_profile)).load(
        bucket_name, object_key
    )


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config1 = {'data_source': DataSource.REDSHIFT}
        config2 = {'data_source': DataSource.S3}
        new_redshift_template = fetch_template_source(BlockType.DATA_LOADER, config1)
        new_s3_template = fetch_template_source(BlockType.DATA_LOADER, config2)
        self.assertEqual(redshift_template, new_redshift_template)
        self.assertEqual(s3_template, new_s3_template)

    def test_template_generation_data_loader_api(self):
        expected_template = """import io
import pandas as pd
import requests
from pandas import DataFrame

if 'data_loader' not in globals():
    from mage_ai.data_preparation.decorators import data_loader
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@data_loader
def load_data_from_api(**kwargs) -> DataFrame:
    \"\"\"
    Template for loading data from API
    \"\"\"
    url = ''

    response = requests.get(url)
    return pd.read_csv(io.StringIO(response.text), sep=',')


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""
        config = {'data_source': DataSource.API}
        api_template = fetch_template_source(BlockType.DATA_LOADER, config)
        self.assertEqual(api_template, expected_template)

    def test_template_generation_transformer_default(self):
        expected_template = """from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def transform_df(df: DataFrame, *args, **kwargs) -> DataFrame:
    \"\"\"
    Template code for a transformer block.

    Add more parameters to this function if this block has multiple parent blocks.
    There should be one parameter for each output variable from each parent block.

    Args:
        df (DataFrame): Data frame from parent block.

    Returns:
        DataFrame: Transformed data frame
    \"\"\"
    # Specify your transformation logic here

    return df


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config1 = {'action_type': 'custom'}
        config2 = {'axis': 'row'}
        config3 = {}
        new_template1 = fetch_template_source(BlockType.TRANSFORMER, config1)
        new_template2 = fetch_template_source(BlockType.TRANSFORMER, config2)
        new_template3 = fetch_template_source(BlockType.TRANSFORMER, config3)
        self.assertEqual(expected_template, new_template1)
        self.assertEqual(expected_template, new_template2)
        self.assertEqual(expected_template, new_template3)

    def test_template_generation_transformer_action_default(self):
        expected_template = """from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def transform_df(df: DataFrame, *args, **kwargs) -> DataFrame:
    \"\"\"
    Template code for a transformer block.

    Add more parameters to this function if this block has multiple parent blocks.
    There should be one parameter for each output variable from each parent block.

    Args:
        df (DataFrame): Data frame from parent block.

    Returns:
        DataFrame: Transformed data frame
    \"\"\"
    # Specify your transformation logic here

    return df


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config1 = {'action_type': 'custom'}
        config2 = {'axis': 'row'}
        config3 = {}
        new_template1 = fetch_template_source(BlockType.TRANSFORMER, config1)
        new_template2 = fetch_template_source(BlockType.TRANSFORMER, config2)
        new_template3 = fetch_template_source(BlockType.TRANSFORMER, config3)
        self.assertEqual(expected_template, new_template1)
        self.assertEqual(expected_template, new_template2)
        self.assertEqual(expected_template, new_template3)

    def test_template_generation_transformer_action_simple(self):
        expected_template = """from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_cleaner.transformer_actions.utils import build_transformer_action
from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def execute_transformer_action(df: DataFrame, *args, **kwargs) -> DataFrame:
    \"\"\"
    Execute Transformer Action: ActionType.CLEAN_COLUMN_NAME

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#clean-column-names
    \"\"\"
    action = build_transformer_action(
        df,
        action_type=ActionType.CLEAN_COLUMN_NAME,
        arguments=df.columns,
        axis=Axis.COLUMN,
    )

    return BaseAction(action).execute(df)


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config = {'action_type': ActionType.CLEAN_COLUMN_NAME, 'axis': Axis.COLUMN}
        new_template = fetch_template_source(BlockType.TRANSFORMER, config)
        self.assertEqual(expected_template, new_template)

    def test_template_generation_transformer_action_action_code(self):
        expected_template = """from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_cleaner.transformer_actions.utils import build_transformer_action
from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def execute_transformer_action(df: DataFrame, *args, **kwargs) -> DataFrame:
    \"\"\"
    Execute Transformer Action: ActionType.FILTER

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#filter
    \"\"\"
    action = build_transformer_action(
        df,
        action_type=ActionType.FILTER,
        axis=Axis.ROW,
        action_code='',  # Specify your filtering code here
    )

    return BaseAction(action).execute(df)


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config = {'action_type': ActionType.FILTER, 'axis': Axis.ROW}
        new_template = fetch_template_source(BlockType.TRANSFORMER, config)
        self.assertEqual(expected_template, new_template)

    def test_template_generation_transformer_action_action_options(self):
        expected_template = """from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_cleaner.transformer_actions.utils import build_transformer_action
from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def execute_transformer_action(df: DataFrame, *args, **kwargs) -> DataFrame:
    \"\"\"
    Execute Transformer Action: ActionType.REFORMAT

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#reformat-values
    \"\"\"
    action = build_transformer_action(
        df,
        action_type=ActionType.REFORMAT,
        arguments=[],  # Specify columns to reformat
        axis=Axis.COLUMN,
        options={'reformat': None},  # Specify reformat action,
    )

    return BaseAction(action).execute(df)


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config = {'action_type': ActionType.REFORMAT, 'axis': Axis.COLUMN}
        new_template = fetch_template_source(BlockType.TRANSFORMER, config)
        self.assertEqual(expected_template, new_template)

    def test_template_generation_transformer_action_outputs(self):
        expected_template = """from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_cleaner.transformer_actions.utils import build_transformer_action
from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def execute_transformer_action(df: DataFrame, *args, **kwargs) -> DataFrame:
    \"\"\"
    Execute Transformer Action: ActionType.FIRST

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#aggregation-actions
    \"\"\"
    action = build_transformer_action(
        df,
        action_type=ActionType.FIRST,
        action_code='',  # Enter filtering condition on rows before aggregation
        arguments=[],  # Enter the columns to compute aggregate over
        axis=Axis.COLUMN,
        options={'groupby_columns': []},  # Enter columns to group by
        outputs=[
            # The number of outputs below must match the number of arguments
            {'uuid': 'new_aggregate_column', 'column_type': 'category'},
        ],
    )

    return BaseAction(action).execute(df)


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config = {'action_type': ActionType.FIRST, 'axis': Axis.COLUMN}
        new_template = fetch_template_source(BlockType.TRANSFORMER, config)
        self.assertEqual(expected_template, new_template)

    def test_template_generation_data_exporter_default(self):
        expected_template = """from pandas import DataFrame

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data(df: DataFrame, **kwargs) -> None:
    \"\"\"
    Exports data to some source

    Args:
        df (DataFrame): Data frame to export to
    \"\"\"
    # Specify your data exporting logic here

"""

        config1 = {'data_source': 'default'}
        config2 = {}
        new_template1 = fetch_template_source(BlockType.DATA_EXPORTER, config1)
        new_template2 = fetch_template_source(BlockType.DATA_EXPORTER, config2)
        self.assertEqual(expected_template, new_template1)
        self.assertEqual(expected_template, new_template2)

    def test_template_generation_data_exporter_specific(self):
        bigquery_template = """from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.bigquery import BigQuery
from mage_ai.io.config import ConfigFileLoader
from pandas import DataFrame
from os import path

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_big_query(df: DataFrame, **kwargs) -> None:
    \"\"\"
    Template for exporting data to a BigQuery warehouse.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/blocks/data_loading.md#bigquery
    \"\"\"
    table_id = 'your-project.your_dataset.your_table_name'
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    BigQuery.with_config(ConfigFileLoader(config_path, config_profile)).export(
        df,
        table_id,
        if_exists='replace',  # Specify resolution policy if table name already exists
    )
"""
        snowflake_template = """from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.snowflake import Snowflake
from pandas import DataFrame
from os import path

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_snowflake(df: DataFrame, **kwargs) -> None:
    \"\"\"
    Template for exporting data to a Snowflake warehouse.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/blocks/data_loading.md#snowflake
    \"\"\"
    table_name = 'your_table_name'
    database = 'your_database_name'
    schema = 'your_schema_name'
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    with Snowflake.with_config(ConfigFileLoader(config_path, config_profile)) as loader:
        loader.export(
            df,
            table_name,
            database,
            schema,
            if_exists='replace',  # Specify resolution policy if table already exists
        )
"""

        config1 = {'data_source': DataSource.BIGQUERY}
        config2 = {'data_source': DataSource.SNOWFLAKE}
        new_bigquery_template = fetch_template_source(BlockType.DATA_EXPORTER, config1)
        new_snowflake_template = fetch_template_source(BlockType.DATA_EXPORTER, config2)
        self.assertEqual(bigquery_template, new_bigquery_template)
        self.assertEqual(snowflake_template, new_snowflake_template)

    def test_template_generation_transformer_dwh(self):
        postgres_template = """from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.postgres import Postgres
from os import path
from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def transform_in_postgres(*args, **kwargs) -> DataFrame:
    \"\"\"
    Performs a transformation in Postgres
    \"\"\"
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    # Specify your SQL transformation query
    query = 'your transformation_query'

    # Specify table to sample data from. Use to visualize changes to table.
    sample_table = 'table_to_sample_data_from'
    sample_schema = 'schema_of_table_to_sample'
    sample_size = 10_000

    with Postgres.with_config(ConfigFileLoader(config_path, config_profile)) as loader:
        # Write queries to transform your dataset with
        loader.execute(query)
        loader.commit() # Permanently apply database changes
        return loader.sample(sample_schema, sample_size, sample_table)


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        bigquery_template = """from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.bigquery import BigQuery
from os import path
from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


@transformer
def transform_in_bigquery(*args, **kwargs) -> DataFrame:
    \"\"\"
    Performs a transformation in BigQuery
    \"\"\"
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    # Specify your SQL transformation query
    query = 'your transformation_query'

    # Specify table to sample data from. Use to visualize changes to table.
    sample_table = 'table_to_sample_data_from'
    sample_schema = 'schema_of_table_to_sample'
    sample_size = 10_000

    with BigQuery.with_config(ConfigFileLoader(config_path, config_profile)) as loader:
        # Write queries to transform your dataset with
        loader.execute(query)
        return loader.sample(sample_schema, sample_size, sample_table)


@test
def test_output(df) -> None:
    \"\"\"
    Template code for testing the output of the block.
    \"\"\"
    assert df is not None, 'The output is undefined'
"""

        config1 = {'data_source': DataSource.POSTGRES}
        config2 = {'data_source': DataSource.BIGQUERY}
        expected_postgres_template = fetch_template_source(BlockType.TRANSFORMER, config1)
        expected_bigquery_template = fetch_template_source(BlockType.TRANSFORMER, config2)
        self.assertEqual(postgres_template, expected_postgres_template)
        self.assertEqual(bigquery_template, expected_bigquery_template)



from mage_ai.services.aws.ecs.config import EcsConfig
import boto3
import json


def run_task(command: str, ecs_config: EcsConfig) -> None:
    if type(ecs_config) is dict:
        ecs_config = EcsConfig.load(config=ecs_config)
    client = boto3.client('ecs')
    response = client.run_task(**ecs_config.get_task_config(command=command))
    print(json.dumps(response, indent=4, default=str))



from mage_ai.data_cleaner.cleaning_rules.base import STATUS_COMPLETED
from mage_ai.data_cleaner.cleaning_rules.clean_column_names import CleanColumnNames
from mage_ai.data_cleaner.cleaning_rules.fix_syntax_errors import FixSyntaxErrors
from mage_ai.data_cleaner.cleaning_rules.impute_values import ImputeValues
from mage_ai.data_cleaner.cleaning_rules.reformat_values import ReformatValues
from mage_ai.data_cleaner.cleaning_rules.remove_collinear_columns import RemoveCollinearColumns
from mage_ai.data_cleaner.cleaning_rules.remove_columns_with_high_empty_rate import (
    RemoveColumnsWithHighEmptyRate,
)
from mage_ai.data_cleaner.cleaning_rules.remove_columns_with_single_value import (
    RemoveColumnsWithSingleValue,
)
from mage_ai.data_cleaner.cleaning_rules.remove_duplicate_rows import RemoveDuplicateRows
from mage_ai.data_cleaner.cleaning_rules.remove_outliers import (
    RemoveOutliers,
    REMOVE_OUTLIERS_TITLE,
)
from mage_ai.data_cleaner.column_types.column_type_detector import infer_column_types
from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.statistics.calculator import StatisticsCalculator
from mage_ai.shared.array import flatten
from mage_ai.shared.constants import SAMPLE_SIZE
from mage_ai.shared.logger import VerboseFunctionExec, timer
import numpy as np

DEFAULT_RULES = [
    CleanColumnNames,
    RemoveColumnsWithHighEmptyRate,
    RemoveColumnsWithSingleValue,
    FixSyntaxErrors,
    ReformatValues,
    ImputeValues,
    RemoveCollinearColumns,
    RemoveDuplicateRows,
    RemoveOutliers,
]


class BasePipeline:
    def __init__(self, actions=[], rules=DEFAULT_RULES, verbose=False):
        self.actions = actions
        self.rules = rules
        self.verbose = verbose

    def create_actions(self, df, column_types, statistics, rule_configs={}):
        if not statistics or len(statistics) == 0:
            calculator = StatisticsCalculator(column_types, self.verbose)
            statistics = calculator.calculate_statistics_overview(df, False)
        self.column_types = column_types
        all_suggestions = []
        for rule in self.rules:
            with timer('pipeline.evaluate_cleaning_rule', dict(rule=rule.__name__), verbose=False):
                with VerboseFunctionExec(
                    f'Evaluating cleaning rule: {rule.__name__}', verbose=self.verbose
                ):
                    rule_config = rule_configs.get(rule.__name__, {})
                    suggestions = \
                        rule(df, column_types, statistics, custom_config=rule_config).evaluate()
            if suggestions:
                all_suggestions += suggestions
        self.actions = all_suggestions
        return all_suggestions

    def create_preview_results(self, df, suggested_actions):
        for action in suggested_actions:
            payload = action['action_payload']
            if payload['axis'] != 'row':
                continue
            df_transformed = BaseAction(payload).execute(df)
            row_removed = df.index.difference(df_transformed.index)
            row_removed = df.index.get_indexer(row_removed)
            row_removed = row_removed[np.where(row_removed <= SAMPLE_SIZE)]
            if len(row_removed) > 0:
                action['preview_results'] = dict(
                    removed_row_indices=row_removed.tolist(),
                )

    def transform(self, df, auto=True):
        if len(self.actions) == 0:
            print('Pipeline is empty.')
            return df
        action_queue = self.actions
        completed_queue = []
        df_transformed = df
        while len(action_queue) != 0:
            action = action_queue.pop(0)
            title = action['title']
            payload = action['action_payload']
            with VerboseFunctionExec(f'Executing cleaning action: {title}', verbose=self.verbose):
                df_transformed = BaseAction(payload).execute(df_transformed)
            action['status'] = STATUS_COMPLETED
            completed_queue.append(action)
            if auto:
                action_queue = self.update_suggestions(df_transformed)
        self.actions = completed_queue
        return df_transformed

    def update_suggestions(self, df_transformed):
        new_column_types = infer_column_types(df_transformed)
        return self.create_actions(df_transformed, new_column_types, {})

    @classmethod
    def deduplicate_suggestions(self, actions, suggestions, statistics):
        """
        Not show duplicate outlier removal suggestions due to column value distribution changes.
        TODO: Figure out a better way to detect outliers.
        """
        columns_with_outlier_removed = set(
            flatten(
                [
                    a['action_payload']['action_arguments']
                    for a in actions
                    if a['title'] == REMOVE_OUTLIERS_TITLE
                ]
            )
        )
        suggestions_filtered = [
            s
            for s in suggestions
            if not (
                s['title'] == REMOVE_OUTLIERS_TITLE
                and len(s['action_payload']['action_arguments']) != 0
                and s['action_payload']['action_arguments'][0] in columns_with_outlier_removed
            )
        ]
        statistics_updated = statistics.copy()
        for col in columns_with_outlier_removed:
            if f'{col}/outlier_count' in statistics_updated:
                statistics_updated[f'{col}/outlier_count'] = 0
                statistics_updated[f'{col}/outliers'] = []
        return suggestions_filtered, statistics_updated



from mage_ai.io.file import FileIO
from pandas import DataFrame

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_file(df: DataFrame, **kwargs) -> None:
    """
    Template for exporting data to filesystem.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/blocks/data_loading.md#fileio
    """
    filepath = 'titanic_clean.csv'
    FileIO().export(df, filepath)



from pyspark.sql import types
from pyspark.sql.types import (
    BooleanType,
    DoubleType,
    IntegerType,
    StringType,
)
import re

COLUMN_TYPE_MAPPING = {
    'category': StringType,
    'category_high_cardinality': StringType,
    'datetime': StringType,
    'email': StringType,
    'number': IntegerType,
    'number_with_decimals': DoubleType,
    'phone_number': StringType,
    'text': StringType,
    'true_or_false': StringType,
    'zip_code': StringType,
}

GROUP_MOD_COLUMN = '__group_mod__'
ROW_NUMBER_COLUMN = '__row_number__'
ROW_NUMBER_LIT_COLUMN = '__row_number_lit__'



from IPython import get_ipython
from IPython.display import IFrame, Javascript, display
from enum import Enum
from mage_ai.server.app import (
    server_config,
)
from mage_ai.server.constants import SERVER_PORT
import logging
import os


FRONTEND_DIST_PATH = os.path.abspath(
    os.path.join(os.path.dirname(os.path.dirname(__file__)), 'frontend_dist'),
)
FRONTEND_STATIC_PATH = '/_next/static'
IFRAME_HEIGHT = 1000

logger = logging.getLogger(__name__)


class NotebookType(str, Enum):
    DATABRICKS = 'databricks'
    GOOGLE_COLAB = 'google_colab'
    SAGEMAKER = 'sagemaker'


def infer_notebook_type():
    if os.environ.get('DATABRICKS_RUNTIME_VERSION'):
        return NotebookType.DATABRICKS
    elif type(get_ipython()).__module__.startswith('google.colab'):
        return NotebookType.GOOGLE_COLAB
    elif os.environ.get('AWS_PATH'):
        return NotebookType.SAGEMAKER
    else:
        return None


def display_inline_iframe(host='localhost', port=SERVER_PORT, notebook_type=None, config={}):
    path_to_server = f'http://{host}:{port}'

    def __print_url():
        print(f'Open UI in another tab with url: {path_to_server}')

    if notebook_type == NotebookType.GOOGLE_COLAB:
        from google.colab.output import eval_js
        path_to_server = eval_js(f'google.colab.kernel.proxyPort({SERVER_PORT})')
        __print_url()
        display(Javascript("""
            (async ()=>{
                fm = document.createElement('iframe')
                fm.src = await google.colab.kernel.proxyPort(%s)
                fm.width = '95%%'
                fm.height = '%d'
                fm.frameBorder = 0
                document.body.append(fm)
            })();
            """ % (SERVER_PORT, IFRAME_HEIGHT)))
    elif notebook_type == NotebookType.DATABRICKS:
        databricks_host = config.get('databricks_host')
        base_path = server_config.server_base_path
        url_params = server_config.server_url_params
        path_to_server = f'https://{databricks_host}{base_path}{url_params}'
        __print_url()
    elif notebook_type == NotebookType.SAGEMAKER:
        sagemaker_host = config.get('sagemaker_host')
        path_to_server = f'https://{sagemaker_host}/proxy/{port}'
        __print_url()
        display(IFrame(path_to_server, width='95%', height=1000))
    else:
        __print_url()
        display(IFrame(path_to_server, width='95%', height=1000))


def update_frontend_urls(host=None, port=None, notebook_type=None, config={}):
    if notebook_type == NotebookType.DATABRICKS:
        required_args = [
            'cluster_id',
            'databricks_host',
            'workspace_id',
            'token',
        ]
        for arg in required_args:
            if arg not in config:
                logger.error(f'Parameter "{arg}" is required to generate proxy url.')
                return
        cluster_id = config.get('cluster_id')
        workspace_id = config.get('workspace_id')
        token = config.get('token')
        base_path = f'/driver-proxy-api/o/{workspace_id}/{cluster_id}/{port}/'
        url_params = f'?token={token}'
        server_config.server_base_path = base_path
        server_config.server_url_params = url_params
    elif notebook_type == NotebookType.SAGEMAKER:
        base_path = f'/proxy/{port}'



from enum import Enum
from mage_ai.shared.utils import clean_name
from pandas import DataFrame, Series
from pandas.api.types import infer_dtype
from typing import Callable, Dict, Mapping

"""
Utilities for exporting Python data frames to external databases.
"""


class BadConversionError(Exception):
    """
    Unable to convert Python-based data type to SQL equivalent.
    """

    pass


class PandasTypes(str, Enum):
    """
    Internal datatypes defined by the pandas Public API
    """

    BOOLEAN = 'boolean'
    BYTES = 'bytes'
    CATEGORICAL = 'categorical'
    COMPLEX = 'complex'
    DATE = 'date'
    DATETIME = 'datetime'
    DATETIME64 = 'datetime64'
    DECIMAL = 'decimal'
    INTEGER = 'integer'
    FLOATING = 'floating'
    MIXED = 'mixed'
    MIXED_INTEGER = 'mixed-integer'
    MIXED_INTEGER_FLOAT = 'mixed-integer-float'
    PERIOD = 'period'
    STRING = 'string'
    TIME = 'time'
    TIMEDELTA = 'timedelta'
    TIMEDELTA64 = 'timedelta64'
    UNKNOWN_ARRAY = 'unknown-array'


def infer_dtypes(df: DataFrame) -> Dict[str, str]:
    """
    Fetches the internal pandas datatypes for the columns in the data frame.

    Args:
        df (DataFrame): Data frame to fetch dtypes from.

    Returns:
        Dict[str, str]: Map of column names to inferred dtypes
    """
    return {column: infer_dtype(df[column], skipna=True) for column in df.columns}


def clean_df_for_export(
    df: DataFrame, column_mapper: Callable[[Series, str], str], dtypes: Mapping[str, str]
) -> DataFrame:
    """
    Cleans data frame with the appropriate steps to prepare loading the data frame
    to the target database.

    Args:
        df (DataFrame): Data frame to clean.
        column_mapper (Callable[[Series, str], str]): Function that cleans a column given the
        pandas data type.
        dtypes (Mapping[str, str]): Name of the new table to create

    Returns:
        str: Table creation query for this table.
    """
    copy_df = df.copy()
    for column in df.columns:
        copy_df[column] = column_mapper(copy_df[column], dtypes[column])
    return copy_df


def gen_table_creation_query(
    dtypes: Mapping[str, str],
    schema_name: str,
    table_name: str,
) -> str:
    """
    Generates a database table creation query from a data frame.

    Args:
        dtypes (Mapping[str, str]): Database relative data types for each column of
        the data frame.
        schema_name (str): Name of schema to create new table in.
        table_name (str): Name of the new table to create.

    Returns:
        str: Table creation query for this table.
    """
    query = []
    for cname in dtypes:
        query.append(f'"{clean_name(cname)}" {dtypes[cname]}')
    return f'CREATE TABLE {schema_name}.{table_name} (' + ','.join(query) + ');'



from contextlib import contextmanager, redirect_stdout
from typing import Callable

import logging
import sys
import time

logger = logging.getLogger(__name__)


class timer(object):
    """
    with timer('metric.metric', tags={ 'key': 'value' }):
        function()
    """

    def __init__(self, metric, tags={}, verbose=True):
        self.metric = metric
        self.start = None
        self.tags = tags
        self.verbose = verbose

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        # Must convert to milliseconds, see details in
        # https://statsd.readthedocs.io/en/v3.1/timing.html
        dt = int((time.time() - self.start) * 1000)
        if self.verbose:
            logger.debug(f'[time] metric: {self.metric}, value: {dt}ms, tags: {self.tags}')


class VerboseFunctionExec:
    def __init__(
        self,
        message: str,
        prefix: str = '',
        verbose: bool = True,
    ):
        self.message = message
        self.verbose = verbose
        self.prefix = prefix

    def __enter__(self):
        if self.verbose:
            enter_message = f'{self.prefix} {self.message}...'
            print(enter_message, end='')

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self.verbose:
            if exc_type is None:
                print(f'{self.prefix} DONE')
            else:
                print(f'{self.prefix} FAILED')


class VerbosePrintHandler:
    def __init__(self, start_msg, verbose=False):
        self.verbose = verbose
        if verbose:
            print(start_msg)
        self.exists_previous_message = False

    @contextmanager
    def print_msg(self, msg):
        if self.verbose:
            if self.exists_previous_message:
                print(f'\r├─ ')
            print(f'└─ {msg}...', end='')
        yield msg
        if self.verbose:
            print('DONE', end='')
            self.exists_previous_message = True


class BlockFunctionExec:
    def __init__(
        self,
        block_uuid: str,
        message: str,
        build_block_output_stdout: Callable[..., object] = None,
    ):
        self.message = message
        self.build_block_output_stdout = build_block_output_stdout
        self.block_uuid = block_uuid

    def __enter__(self):
        if self.build_block_output_stdout is not None:
            stdout = self.build_block_output_stdout(self.block_uuid)
            with redirect_stdout(stdout):
                print(self.message)

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self.build_block_output_stdout is not None:
            stdout = self.build_block_output_stdout(
                self.block_uuid,
                execution_state='idle',
            )
            with redirect_stdout(stdout):
                print('DONE')



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.REMOVE_OUTLIERS

    Warning: This method uses relative outlier checks, and so repeated executions of this
    transformer action will continue to remove data.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#remove-outliers
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.REMOVE_OUTLIERS,
        arguments=df.columns,  # Specify columns to remove outliers from
        axis=Axis.COLUMN,
        options={'method': 'auto'},  # Specify algorithm to use for outlier removal
    )
{% endblock %}



from mage_ai.io.base import BaseSQLConnection, ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from pandas import DataFrame
from snowflake.connector import connect
from snowflake.connector.pandas_tools import write_pandas

DEFAULT_LOGIN_TIMEOUT = 20
# NOTE: if credentials are wrong, it’ll take this many seconds for the user to be shown an error.
# TODO: check credentials before executing query and error sooner
# than waiting for Snowflake to timeout.
DEFAULT_NETWORK_TIMEOUT = 60 * 5


class Snowflake(BaseSQLConnection):
    """
    Handles data transfer between a Snowflake data warehouse and the Mage app.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initializes settings for connecting to Snowflake data warehouse.
        The following arguments must be provided to the connector, all other
        arguments are optional.

        Required Arguments:
            user (str): Username for the Snowflake user.
            password (str): Login Password for the user.
            account (str): Snowflake account identifier (excluding
            `snowflake-computing.com` suffix).
        """
        if 'login_timeout' not in kwargs:
            kwargs['login_timeout'] = DEFAULT_LOGIN_TIMEOUT
        if 'network_timeout' not in kwargs:
            kwargs['network_timeout'] = DEFAULT_NETWORK_TIMEOUT
        if kwargs.get('verbose') is not None:
            kwargs.pop('verbose')
        super().__init__(verbose=kwargs.get('verbose', True), **kwargs)

    def open(self) -> None:
        """
        Opens a connection to Snowflake.
        """
        with self.printer.print_msg('Connecting to Snowflake warehouse'):
            self._ctx = connect(**self.settings)

    def execute(self, query_string: str, **kwargs) -> None:
        """
        Executes any query in the Snowflake data warehouse.

        Args:
            query_string (str): The query to execute on Snowflake's platform.
            **kwargs: Additional parameters to provide to the query
        """
        with self.printer.print_msg(f'Executing query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            with self.conn.cursor() as cur:
                return cur.execute(query_string, **kwargs).fetchall()

    def load(
        self,
        query_string: str,
        limit: int = QUERY_ROW_LIMIT,
        display_query: str = None,
        verbose: bool = True,
        *args,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from Snowflake into a Pandas data frame based on the query given.
        This will fail unless a `SELECT` query is provided. This function will load at
        maximum 100,000 rows of data. To operate on more data, consider performing data
        transformations in warehouse.


        Args:
            query_string (str): Query to fetch a table or subset of a table.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to 100000.
            *args, **kwargs: Additional parameters to provide to the query

        Returns:
            DataFrame: Data frame associated with the given query.
        """

        print_message = 'Loading data'
        if verbose:
            print_message += ' with query'

            if display_query:
                for line in display_query.split('\n'):
                    print_message += f'\n{line}'
            else:
                print_message += f'\n{query_string}'

        query_string = self._clean_query(query_string)

        with self.printer.print_msg(print_message):
            with self.conn.cursor() as cur:
                return cur.execute(
                    self._enforce_limit(query_string, limit), *args, **kwargs
                ).fetch_pandas_all()

    def export(
        self,
        df: DataFrame,
        table_name: str,
        database: str,
        schema: str,
        if_exists: str = 'append',
        query_string: str = None,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        """
        Exports a Pandas data frame to a Snowflake warehouse based on the table name. If table doesn't
        exist, the table is automatically created.

        Args:
            df (DataFrame): Data frame to export to a Snowflake warehouse.
            table_name (str): Name of the table to export data to (excluding database and schema).
            database (str): Name of the database in which the table is located.
            schema (str): Name of the schema in which the table is located.
            if_exists (str, optional): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table.
            Defaults to `'append'`.
            **kwargs: Additional arguments to pass to writer
        """

        def __process():
            with self._ctx.cursor() as cur:
                cur.execute(f'SHOW TABLES LIKE \'{table_name}\' IN SCHEMA {database}.{schema}')

                table_doesnt_exist = cur.rowcount == 1
                if cur.rowcount > 1:
                    raise ValueError(f'Two or more tables with the name {table_name} are found.')
                elif not table_doesnt_exist:
                    if ExportWritePolicy.FAIL == if_exists:
                        raise RuntimeError(
                            f'Table {table_name} already exists in the current warehouse, database, schema scenario.'
                        )
                    elif ExportWritePolicy.REPLACE == if_exists:
                        cur.execute(f'USE DATABASE {database}')
                        cur.execute(f'DROP TABLE "{schema}"."{table_name}"')

                if query_string:
                    if table_doesnt_exist:
                        cur.execute(f'USE DATABASE {database}')
                        cur.execute(f"""
CREATE TABLE IF NOT EXISTS "{database}"."{schema}"."{table_name}" AS
{query_string}
""")
                    elif not table_doesnt_exist:
                        cur.execute(f'USE DATABASE {database}')
                        cur.execute(f"""
INSERT INTO "{database}"."{schema}"."{table_name}"
{query_string}
""")

            auto_create_table = True
            if 'auto_create_table' in kwargs:
                auto_create_table = kwargs.pop(auto_create_table)
                if auto_create_table is None:
                    auto_create_table = True

            if not query_string:
                write_pandas(
                    self.conn,
                    df,
                    table_name,
                    database=database,
                    schema=schema,
                    auto_create_table=auto_create_table,
                    **kwargs,
                )

        if verbose:
            with self.printer.print_msg(
                f'Exporting data to \'{database}.{schema}.{table_name}\''
            ):
                __process()
        else:
            __process()

    @classmethod
    def with_config(
        cls,
        config: BaseConfigLoader,
        database=None,
        schema=None,
        **kwargs,
    ) -> 'Snowflake':
        """
        Initializes Snowflake client from configuration loader.

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        return cls(
            user=config[ConfigKey.SNOWFLAKE_USER],
            password=config[ConfigKey.SNOWFLAKE_PASSWORD],
            account=config[ConfigKey.SNOWFLAKE_ACCOUNT],
            warehouse=config[ConfigKey.SNOWFLAKE_DEFAULT_WH],
            database=database or config[ConfigKey.SNOWFLAKE_DEFAULT_DB],
            schema=schema or config[ConfigKey.SNOWFLAKE_DEFAULT_SCHEMA],
            **kwargs,
        )



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.base import DataSource
from mage_ai.server.api.base import BaseHandler
import yaml


DATA_PROVIDERS = [
    DataSource.BIGQUERY,
    DataSource.POSTGRES,
    DataSource.REDSHIFT,
    DataSource.SNOWFLAKE,
]


class ApiDataProvidersHandler(BaseHandler):
    def get(self):
        profiles = []

        with open(f'{get_repo_path()}/io_config.yaml', 'r') as stream:
            try:
                profiles = list(yaml.safe_load(stream).keys())
            except yaml.YAMLError as exc:
                print(exc)

        collection = [dict(
            id=ds.title(),
            profiles=[p for p in profiles if p != 'version'],
            value=ds.value,
        ) for ds in DATA_PROVIDERS]
        self.write(dict(data_providers=collection))



from faker import Faker
from mage_ai.data_cleaner.column_types.column_type_detector import infer_column_types
from mage_ai.data_cleaner.statistics.calculator import StatisticsCalculator
from mage_ai.tests.base_test import TestCase
from random import shuffle
import pandas as pd
import numpy as np


class StatisticsCalculatorTest(TestCase):
    def setUp(self):
        self.faker = Faker(locale='en_US')
        return super().setUp()

    def test_calculate_statistics_overview(self):
        calculator = StatisticsCalculator(
            column_types=dict(
                age='number',
                country='category',
                date_joined='datetime',
                id='number',
                cancelled='true_or_false',
            ),
        )

        df = pd.DataFrame(
            [
                [1, 'JP', '2000-01-01', '1', False],
                [2, 'KO', '2000-12-01', '2', False],
                [1, 'US', '2000-07-01 00:00:00+00:00', '3', False],
                [' ', 'US', '2000-07-01 00:00:00+00:00', '4', False],
                [1, None, '2000-07-01 00:00:00+00:00', '5', False],
                ['', None, '899-07-01 00:00:00+00:00', '6', False],
                [3, 'US', None, '7', True],
            ],
            columns=['age', 'country', 'date_joined', 'id', 'cancelled'],
        )

        data = calculator.calculate_statistics_overview(df, is_clean=False)

        self.assertEqual(data['count'], 7)
        self.assertEqual(data['total_null_value_count'], 6)

        self.assertEqual(data['age/average'], 8 / 5)
        self.assertEqual(data['age/count'], 5)
        self.assertEqual(data['age/count_distinct'], 3)
        self.assertEqual(data['age/max'], 3)
        self.assertEqual(data['age/median'], 1)
        self.assertEqual(data['age/min'], 1)
        self.assertEqual(data['age/sum'], 8)
        self.assertEqual(data['age/null_value_rate'], 2 / 7)
        self.assertEqual(data['age/completeness'], 1 - data['age/null_value_rate'])
        self.assertEqual(data['age/quality'], 'Bad')

        self.assertEqual(data['country/count'], 5)
        self.assertEqual(data['country/count_distinct'], 3)
        self.assertEqual(data['country/mode'], 'US')
        self.assertEqual(data['country/null_value_rate'], 2 / 7)
        self.assertEqual(data['country/completeness'], 1 - data['country/null_value_rate'])
        self.assertEqual(data['country/quality'], 'Bad')

        self.assertEqual(data['date_joined/count'], 5)
        self.assertEqual(data['date_joined/count_distinct'], 3)
        self.assertEqual(data['date_joined/max'], '2000-12-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/median'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/min'], '2000-01-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/mode'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/null_value_rate'], 2 / 7)
        self.assertEqual(data['date_joined/completeness'], 1 - data['date_joined/null_value_rate'])
        self.assertEqual(data['date_joined/quality'], 'Bad')

        self.assertEqual(data['id/count'], 7)
        self.assertEqual(data['id/count_distinct'], 7)
        self.assertEqual(data['id/null_value_rate'], 0)
        self.assertEqual(data['id/completeness'], 1)
        self.assertEqual(data['id/quality'], 'Good')

        self.assertEqual(data['cancelled/mode'], False)
        self.assertEqual(data['cancelled/null_value_rate'], 0)
        self.assertEqual(data['cancelled/completeness'], 1)
        self.assertEqual(data['cancelled/quality'], 'Good')

    def test_calculate_statistics_overview_invalid_indices(self):
        df = pd.DataFrame(
            [
                [1, 'abc@xyz.com', 32132],
                [2, None, 12345],
                [3, 'test', 1234],
                [4, 'abc@test.net', 'abcde'],
                [5, 'abc12345@', 54321],
                [6, 'abcdef@123.com', 56789],
            ],
            columns=['id', 'email', 'zip_code'],
        )
        calculator = StatisticsCalculator(
            column_types=dict(
                id='number',
                email='email',
                zip_code='zip_code',
            ),
        )
        data = calculator.calculate_statistics_overview(df, is_clean=True)

        self.assertEqual(data['email/invalid_values'], ['test', 'abc12345@'])
        self.assertTrue((data['email/invalid_indices'] == np.array([2, 4])).all())
        self.assertEqual(data['email/invalid_value_count'], 2)
        self.assertEqual(data['email/invalid_value_rate'], 2 / 6)

        self.assertEqual(data['zip_code/invalid_values'], ['abcde'])
        self.assertTrue((data['zip_code/invalid_indices'] == np.array([3])).all())
        self.assertEqual(data['zip_code/invalid_value_count'], 1)
        self.assertEqual(data['zip_code/invalid_value_rate'], 1 / 6)

        self.assertEqual(data['id/invalid_values'], [])
        self.assertTrue((data['id/invalid_indices'] == np.array([]).astype(int)).all())
        self.assertEqual(data['id/invalid_value_count'], 0)
        self.assertEqual(data['id/invalid_value_rate'], 0 / 6)

    def test_calculate_statistics_overview_no_cleaning(self):
        calculator = StatisticsCalculator(
            column_types=dict(
                age='number',
                country='category',
                date_joined='datetime',
                id='number',
                cancelled='true_or_false',
            ),
        )

        df = pd.DataFrame(
            [
                [1, 'JP', '2000-01-01', 1, False],
                [2, 'KO', '2000-12-01', 2, False],
                [1, 'US', '2000-07-01 00:00:00+00:00', 3, False],
                [np.nan, 'US', '2000-07-01 00:00:00+00:00', 4, False],
                [1, None, '2000-07-01 00:00:00+00:00', 5, False],
                [np.nan, None, '899-07-01 00:00:00+00:00', 6, False],
                [3, 'US', None, 7, True],
            ],
            columns=['age', 'country', 'date_joined', 'id', 'cancelled'],
        )

        data = calculator.calculate_statistics_overview(df, is_clean=True)

        self.assertEqual(data['count'], 7)
        self.assertEqual(data['total_null_value_count'], 5)

        self.assertEqual(data['age/average'], 8 / 5)
        self.assertEqual(data['age/count'], 5)
        self.assertEqual(data['age/count_distinct'], 3)
        self.assertEqual(data['age/max'], 3)
        self.assertEqual(data['age/median'], 1)
        self.assertEqual(data['age/min'], 1)
        self.assertEqual(data['age/sum'], 8)
        self.assertEqual(data['age/null_value_rate'], 2 / 7)
        self.assertEqual(data['age/completeness'], 1 - data['age/null_value_rate'])
        self.assertEqual(data['age/quality'], 'Bad')

        self.assertEqual(data['country/count'], 5)
        self.assertEqual(data['country/count_distinct'], 3)
        self.assertEqual(data['country/mode'], 'US')
        self.assertEqual(data['country/null_value_rate'], 2 / 7)
        self.assertEqual(data['country/completeness'], 1 - data['country/null_value_rate'])
        self.assertEqual(data['country/quality'], 'Bad')

        self.assertEqual(data['date_joined/count'], 6)
        self.assertEqual(data['date_joined/count_distinct'], 4)
        self.assertEqual(data['date_joined/max'], '2000-12-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/median'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/min'], '2000-01-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/mode'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/null_value_rate'], 1 / 7)
        self.assertEqual(data['date_joined/completeness'], 1 - data['date_joined/null_value_rate'])
        self.assertEqual(data['date_joined/quality'], 'Good')

        self.assertEqual(data['id/count'], 7)
        self.assertEqual(data['id/count_distinct'], 7)
        self.assertEqual(data['id/null_value_rate'], 0)
        self.assertEqual(data['id/completeness'], 1)
        self.assertEqual(data['id/quality'], 'Good')

        self.assertEqual(data['cancelled/mode'], False)
        self.assertEqual(data['cancelled/null_value_rate'], 0)
        self.assertEqual(data['cancelled/completeness'], 1)
        self.assertEqual(data['cancelled/quality'], 'Good')

    def test_calculate_statistics_overview_divide_by_zero(self):
        calculator = StatisticsCalculator(
            column_types=dict(age='number', date='datetime', category='category', string='string'),
        )

        df = pd.DataFrame([], columns=['age', 'date', 'category', 'string'])
        data = calculator.calculate_statistics_overview(df, is_clean=False)

        self.assertEqual(data['age/count'], 0)
        self.assertEqual(data['age/count_distinct'], 0)
        self.assertEqual(data['age/null_value_rate'], 0)
        self.assertEqual(data['age/unique_value_rate'], 0)
        self.assertEqual(data['age/invalid_value_rate'], 0)
        self.assertEqual(data['date/count'], 0)
        self.assertEqual(data['date/count_distinct'], 0)
        self.assertEqual(data['date/null_value_rate'], 0)
        self.assertEqual(data['date/unique_value_rate'], 0)
        self.assertEqual(data['date/invalid_value_rate'], 0)
        self.assertEqual(data['category/count'], 0)
        self.assertEqual(data['category/count_distinct'], 0)
        self.assertEqual(data['category/null_value_rate'], 0)
        self.assertEqual(data['category/unique_value_rate'], 0)
        self.assertEqual(data['category/invalid_value_rate'], 0)
        self.assertEqual(data['string/count'], 0)
        self.assertEqual(data['string/count_distinct'], 0)
        self.assertEqual(data['string/null_value_rate'], 0)
        self.assertEqual(data['string/unique_value_rate'], 0)
        self.assertEqual(data['string/invalid_value_rate'], 0)
        self.assertEqual(data['total_null_value_rate'], 0)
        self.assertEqual(data['total_invalid_value_rate'], 0)
        self.assertEqual(data['duplicate_row_rate'], 0)

    def test_calculate_statistics_handle_nulls(self):
        calculator = StatisticsCalculator(
            column_types=dict(
                age='number',
                country='category',
                date_joined='datetime',
                id='number',
                cancelled='true_or_false',
            ),
        )

        df = pd.DataFrame(
            [
                [1, None, None, 1, False],
                [np.nan, None, '2000-12-01', 2, False],
                [np.nan, 'US', '2000-07-01 00:00:00+00:00', 3, False],
                [np.nan, 'US', '2000-07-01 00:00:00+00:00', 4, False],
                [1, None, None, 5, False],
                [np.nan, None, '899-07-01 00:00:00+00:00', 6, False],
                [3, 'JP', None, 7, True],
            ],
            columns=['age', 'country', 'date_joined', 'id', 'cancelled'],
        )

        data = calculator.calculate_statistics_overview(df, is_clean=True)

        self.assertEqual(data['count'], 7)
        self.assertEqual(data['total_null_value_count'], 11)

        self.assertEqual(data['age/average'], 5 / 3)
        self.assertEqual(data['age/count'], 3)
        self.assertEqual(data['age/count_distinct'], 2)
        self.assertEqual(data['age/max'], 3)
        self.assertEqual(data['age/median'], 1)
        self.assertEqual(data['age/min'], 1)
        self.assertEqual(data['age/sum'], 5)
        self.assertEqual(data['age/null_value_rate'], 4 / 7)
        self.assertEqual(data['age/completeness'], 1 - data['age/null_value_rate'])
        self.assertEqual(data['age/quality'], 'Bad')

        self.assertEqual(data['country/count'], 3)
        self.assertEqual(data['country/count_distinct'], 2)
        self.assertEqual(data['country/mode'], 'US')
        self.assertEqual(data['country/null_value_rate'], 4 / 7)
        self.assertEqual(data['country/completeness'], 1 - data['country/null_value_rate'])
        self.assertEqual(data['country/quality'], 'Bad')

        self.assertEqual(data['date_joined/count'], 4)
        self.assertEqual(data['date_joined/count_distinct'], 3)
        self.assertEqual(data['date_joined/max'], '2000-12-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/median'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/min'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/mode'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/null_value_rate'], 3 / 7)
        self.assertEqual(data['date_joined/completeness'], 1 - data['date_joined/null_value_rate'])
        self.assertEqual(data['date_joined/quality'], 'Bad')

        self.assertEqual(data['id/count'], 7)
        self.assertEqual(data['id/count_distinct'], 7)
        self.assertEqual(data['id/null_value_rate'], 0)
        self.assertEqual(data['id/completeness'], 1)
        self.assertEqual(data['id/quality'], 'Good')

        self.assertEqual(data['cancelled/mode'], False)
        self.assertEqual(data['cancelled/null_value_rate'], 0)
        self.assertEqual(data['cancelled/completeness'], 1)
        self.assertEqual(data['cancelled/quality'], 'Good')

    def test_calculate_statistics_text_data(self):
        texts = [
            'Know fine seat prevent 92383ee3. \nCreate Mr. real',
            'Avoid seat place.\nTrial exist against create.',
            'Exactly question mention floor time. Tree theory central seat important beyond hour.',
            'Physical father different ago away place. Health enough product even goal seat team.',
            None,
            None,
            'I know that bees make honey, but do they bake cake?',
            'It is not a problem with the quantum flux capacitor, but the light conduits',
            'Breaking News: scientists finally figure out if water is wet, leading to widespread controversy',
            'The Mage data cleaning tool is overpowered, try it out now!',
            None,
            None,
        ]
        shuffle(texts)
        df = pd.DataFrame({'text': texts})
        calculator = StatisticsCalculator(
            column_types=dict(
                text='text',
            ),
        )
        data = calculator.calculate_statistics_overview(df, is_clean=True)
        self.assertEquals(data['text/word_distribution']['seat'], 4)
        self.assertEquals(data['text/word_distribution']['create'], 2)
        self.assertEquals(data['text/word_distribution']['place'], 2)
        self.assertEquals(data['text/max_word_count'], 14)
        self.assertEquals(data['text/min_word_count'], 7)

    def test_calculate_statistics_email_data(self):
        # This data was generated using Faker
        emails = [
            'xclarke@yahoo.com',
            'daniel40@gmail.com',
            'not an email',
            'phammary@yahoo.com',
            'welchjeffrey@gmail.com',
            None,
            'an improperly formatted email@net.co',
            'wardcindy@gmail.com',
            'bdavis@yahoo.com',
            'qcarpenter@gmail.com',
            'email@net@net.com',
            'larsonkatherine@hotmail.com',
            None,
            'edwardswilliam@jones.org',
            'ojensen@hotmail.com',
            'karenmcbride@gmail.com',
            'danielharris@kim.com',
            'not an email',
        ]
        shuffle(emails)
        df = pd.DataFrame({'emails': emails})
        calculator = StatisticsCalculator(
            column_types=dict(
                emails='email',
            ),
        )
        data = calculator.calculate_statistics_overview(df, is_clean=True)

        expected_domain_distribution = {
            'gmail.com': 5,
            'yahoo.com': 3,
            'hotmail.com': 2,
            'jones.org': 1,
            'kim.com': 1,
        }

        self.assertEquals(expected_domain_distribution, data['emails/domain_distribution'])

    def test_calculate_statistics_list_data(self):
        lists = [
            [2, 'string', False, None],
            None,
            [np.nan, 2.0, 'string', '3'],
            ['not string?', True, True, 8, False, np.nan, None],
            "['not string?'     ,  True,True   ,8, False, np.nan, None]",
            [],
            [8, 9, 9, 8, 'pop', 'string'],
        ]
        df = pd.DataFrame({'lists': lists})
        column_types = infer_column_types(df)
        calculator = StatisticsCalculator(column_types)
        data = calculator.calculate_statistics_overview(df, is_clean=False)
        self.assertTrue(data['lists/most_frequent_element'] is np.nan)
        self.assertEqual(data['lists/least_frequent_element'], 'pop')
        self.assertEqual(data['lists/max_list_length'], 7)
        self.assertEqual(data['lists/min_list_length'], 0)
        self.assertAlmostEqual(data['lists/avg_list_length'], 4.6667, 3)

        expected_list_length_distribution = {
            4: 2,
            7: 2,
            6: 1,
            0: 1,
        }
        self.assertEquals(expected_list_length_distribution, data['lists/length_distribution'])

        expected_element_distribution = {
            'string': 3,
            8: 4,
            True: 4,
            False: 3,
            9: 2,
            np.nan: 7,
            'pop': 1,
            'not string?': 2,
            2: 2,
            '3': 1,
        }
        self.assertEquals(expected_element_distribution, data['lists/element_distribution'])

    def test_calculate_statistics_list_data_with_dictionaries(self):
        lists = [
            [{'e': 2}, 'string', False, None],
            None,
            [np.nan, 2.0, 'string', '3'],
            ['not string?', True, True, 8, False, np.nan, None, {'another_dictionary': -2342.21}],
            "['not string?'     ,  True,True   ,8, False, np.nan, None]",
            [{'dictionary': 'weird'}],
            [8, 9, 9, 8, 'pop', 'string'],
        ]
        df = pd.DataFrame({'lists': lists})
        column_types = infer_column_types(df)
        calculator = StatisticsCalculator(column_types)
        data = calculator.calculate_statistics_overview(df, is_clean=False)
        self.assertTrue(data['lists/most_frequent_element'] is np.nan)
        self.assertEqual(data['lists/least_frequent_element'], 'pop')
        self.assertEqual(data['lists/max_list_length'], 8)
        self.assertEqual(data['lists/min_list_length'], 1)
        self.assertAlmostEqual(data['lists/avg_list_length'], 5, 3)

        expected_list_length_distribution = {
            4: 2,
            7: 1,
            6: 1,
            1: 1,
            8: 1,
        }
        self.assertEquals(expected_list_length_distribution, data['lists/length_distribution'])

    def test_calculate_statistics_box_plot(self):
        df = pd.DataFrame(
            [
                [1000, 'US', 30000, 10, 'cute animal #1', 100, 30],
                [500, 'CA', 10000000, 20, 'intro to regression', 3000, 20],
                [200, '', np.nan, 50, 'daily news #1', None, 75],
                [250, 'CA', 7500, 25, 'machine learning seminar', 8000, 20],
                [1000, 'MX', 45003, 20, 'cute animal #4', 90, 40],
                [1500, 'MX', 75000, 30, '', 70, 25],
                [1500, 'US', 75000, np.nan, 'daily news #3', 70, 25],
                [None, 'US', 75000, 30, 'tutorial: how to start a startup', 70, np.nan],
                [1250, 'US', 60000, 50, 'cute animal #3', 80, 20],
                [200, 'CA', 5000, 30, '', 10000, 30],
                [800, 'US', 50, 40, 'meme compilation', 2000, 45],
                [600, 'CA', 11000, 50, 'daily news #2', 3000, 50],
                [600, 'CA', '', 50, '', 3000, None],
                [700, 'MX', 11750, 20, 'cute animal #2', 2750, 55],
                [700, '', None, 20, '', None, 55],
                [700, 'MX', 11750, '', '', 2750, 55],
                [1200, 'MX', 52000, 10, 'vc funding strats', 75, 60],
            ],
            columns=[
                'number_of_users',
                'location',
                'views',
                'number_of_creators',
                'name',
                'losses',
                'number_of_advertisers',
            ],
        )

        column_types = {
            'number_of_users': 'number',
            'location': 'category',
            'views': 'number',
            'number_of_creators': 'number',
            'name': 'text',
            'losses': 'number',
            'number_of_advertisers': 'number',
        }
        calculator = StatisticsCalculator(column_types=column_types)
        data = calculator.calculate_statistics_overview(df, is_clean=False)

        expected_box_plot_no_users = {
            'outliers': [],
            'min': 200.0,
            'first_quartile': 600.0,
            'median': 700.0,
            'third_quartile': 1000.0,
            'max': 1500.0,
        }

        expected_box_plot_views = {
            'outliers': [10000000.0],
            'min': 50.0,
            'first_quartile': 11000.0,
            'median': 37501.5,
            'third_quartile': 75000.0,
            'max': 75000.0,
        }

        self.assertEqual(expected_box_plot_no_users, data['number_of_users/box_plot_data'])
        self.assertEqual(expected_box_plot_views, data['views/box_plot_data'])



from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.data_preparation.executors.pipeline_executor import PipelineExecutor
from mage_ai.data_preparation.models.constants import ExecutorType, PipelineType
from mage_ai.data_preparation.models.pipeline import Pipeline


class ExecutorFactory:
    @classmethod
    def get_pipeline_executor(
        self,
        pipeline: Pipeline,
        executor_type: ExecutorType = None
    ) -> PipelineExecutor:
        if executor_type is None:
            if pipeline.type == PipelineType.PYSPARK:
                executor_type = ExecutorType.PYSPARK
            else:
                executor_type = ExecutorType.LOCAL_PYTHON
        if executor_type == ExecutorType.PYSPARK:
            from mage_ai.data_preparation.executors.pyspark_pipeline_executor \
                import PySparkPipelineExecutor
            return PySparkPipelineExecutor(pipeline)
        else:
            return PipelineExecutor(pipeline)

    @classmethod
    def get_block_executor(
        self,
        pipeline: Pipeline,
        block_uuid: str,
        execution_partition: str = None,
        executor_type: ExecutorType = None,
    ) -> BlockExecutor:
        executor_kwargs = dict(
            pipeline=pipeline,
            block_uuid=block_uuid,
            execution_partition=execution_partition,
        )
        if executor_type is None:
            if pipeline.type == PipelineType.PYSPARK:
                executor_type = ExecutorType.PYSPARK
            else:
                executor_type = pipeline.get_block(block_uuid).executor_type
        if executor_type == ExecutorType.PYSPARK:
            from mage_ai.data_preparation.executors.pyspark_block_executor \
                import PySparkBlockExecutor
            return PySparkBlockExecutor(**executor_kwargs)
        elif executor_type == ExecutorType.ECS:
            from mage_ai.data_preparation.executors.ecs_block_executor \
                import EcsBlockExecutor
            return EcsBlockExecutor(**executor_kwargs)
        elif executor_type == ExecutorType.K8S:
            from mage_ai.data_preparation.executors.k8s_block_executor \
                import K8sBlockExecutor
            return K8sBlockExecutor(**executor_kwargs)
        else:
            return BlockExecutor(**executor_kwargs)



from dataclasses import dataclass
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.strings import camel_to_snake_case
from typing import Dict
import os
import yaml


@dataclass
class BaseConfig:

    @classmethod
    def load(self, config_path: str = None, config: Dict = None):
        config_class_name = self.__name__
        config_class_key = camel_to_snake_case(config_class_name)
        if config is None:
            self.config_path = config_path
            if self.config_path is None:
                raise Exception(
                    'Please provide a config_path or a config dictionary to initialize'
                    f' an {config_class_name} object',
                )
            if not os.path.exists(self.config_path):
                raise Exception(f'{config_class_name} {self.config_path} does not exist.')
            with open(self.config_path) as fp:
                config = yaml.full_load(fp) or {}

        if config_class_key in config:
            config = config[config_class_key]
        extra_config = self.load_extra_config()
        config = merge_dict(config, extra_config)
        return self(**config)

    @classmethod
    def load_extra_config(self):
        return dict()



from mage_ai.orchestration.notification.config import NotificationConfig
from mage_ai.services.email.email import send_email
from mage_ai.services.slack.slack import send_slack_message


class NotificationSender:
    def __init__(self, config: NotificationConfig):
        self.config = config

    def send(self, message: str = None, subject: str = None) -> None:
        if message is None:
            return
        if self.config.slack_config is not None and self.config.slack_config.is_valid:
            send_slack_message(self.config.slack_config, message)
        if self.config.email_config is not None and subject is not None:
            send_email(self.config.email_config, subject=subject, message=message)



from mage_ai.data_cleaner.column_types.constants import ColumnType
from enum import Enum
import pandas as pd
import numpy as np
import re


CONSTANT_IMPUTATION_DEFAULTS = {
    ColumnType.CATEGORY: 'missing',
    ColumnType.CATEGORY_HIGH_CARDINALITY: 'missing',
    ColumnType.DATETIME: pd.Timestamp.min,
    ColumnType.EMAIL: 'missing',
    ColumnType.LIST: '[]',
    ColumnType.NUMBER: 0,
    ColumnType.NUMBER_WITH_DECIMALS: 0,
    ColumnType.TEXT: 'missing',
    ColumnType.TRUE_OR_FALSE: 'missing',
    ColumnType.PHONE_NUMBER: 'missing',
    ColumnType.ZIP_CODE: 'missing',
}
CURRENCY_SYMBOLS = re.compile(r'(?:[\$\€\¥\₹\元\£]|(?:Rs)|(?:CAD))')
INVALID_VALUE_PLACEHOLDERS = {
    ColumnType.CATEGORY: 'invalid',
    ColumnType.CATEGORY_HIGH_CARDINALITY: 'invalid',
    ColumnType.DATETIME: pd.NaT,
    ColumnType.EMAIL: 'invalid',
    ColumnType.LIST: 'invalid',
    ColumnType.NUMBER: np.nan,
    ColumnType.NUMBER_WITH_DECIMALS: np.nan,
    ColumnType.TEXT: 'invalid',
    ColumnType.TRUE_OR_FALSE: 'invalid',
    ColumnType.PHONE_NUMBER: 'invalid',
    ColumnType.ZIP_CODE: 'invalid',
}


class ActionType(str, Enum):
    ADD = 'add'
    AVERAGE = 'average'
    CLEAN_COLUMN_NAME = 'clean_column_name'
    COUNT = 'count'
    COUNT_DISTINCT = 'count_distinct'
    CUSTOM = 'custom'
    DIFF = 'diff'
    DROP_DUPLICATE = 'drop_duplicate'
    EXPAND_COLUMN = 'expand_column'
    EXPLODE = 'explode'
    FILTER = 'filter'
    FIRST = 'first'
    FIX_SYNTAX_ERRORS = 'fix_syntax_errors'
    GROUP = 'group'
    IMPUTE = 'impute'
    JOIN = 'join'
    LAST = 'last'
    LIMIT = 'limit'
    MAX = 'max'
    MEDIAN = 'median'
    MIN = 'min'
    MODE = 'mode'
    REFORMAT = 'reformat'
    REMOVE = 'remove'
    REMOVE_OUTLIERS = 'remove_outliers'
    SCALE = 'scale'
    SELECT = 'select'
    SHIFT_DOWN = 'shift_down'
    SHIFT_UP = 'shift_up'
    SORT = 'sort'
    SUM = 'sum'
    UNION = 'union'
    UPDATE_TYPE = 'update_type'
    UPDATE_VALUE = 'update_value'


class Axis(str, Enum):
    COLUMN = 'column'
    ROW = 'row'


class VariableType(str, Enum):
    FEATURE = 'feature'
    FEATURE_SET = 'feature_set'
    FEATURE_SET_VERSION = 'feature_set_version'


class Operator(str, Enum):
    CONTAINS = 'contains'
    NOT_CONTAINS = 'not contains'
    EQUALS = '=='
    NOT_EQUALS = '!='
    GREATER_THAN = '>'
    GREATER_THAN_OR_EQUALS = '>='
    LESS_THAN = '<'
    LESS_THAN_OR_EQUALS = '<='


class ImputationStrategy(str, Enum):
    AVERAGE = 'average'
    COLUMN = 'column'
    CONSTANT = 'constant'
    MEDIAN = 'median'
    MODE = 'mode'
    NOOP = 'no_action'
    RANDOM = 'random'
    ROW_RM = 'remove_rows'
    SEQ = 'sequential'


class NameConventionPatterns:
    SNAKE = re.compile(r'^[a-z]+(?:\_[a-z0-9]+)*$')
    CAMEL = re.compile(r'^[a-z]+(?:[A-Z][a-z]*)*$')
    CAMEL_COMPONENT = re.compile(r'^[a-z]+')
    PASCAL = re.compile(r'^(?:[A-Z][a-z]*)+$')
    PASCAL_COMPONENT = re.compile(r'[A-Z][a-z]*')
    UPPERCASE = re.compile(r'^[A-Z0-9]+$')
    NON_ALNUM = re.compile(r'[^a-zA-Z\_0-9]')
    CONNECTORS = re.compile(r'[\s\-\.]')
    LOWERCASE = re.compile(r'^[a-z0-9\_]+$')



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.column_types.constants import (
    CATEGORICAL_TYPES,
    NUMBER_TYPES,
    STRING_TYPES,
    ColumnType,
)
from mage_ai.data_cleaner.shared.utils import wrap_column_name
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
    ImputationStrategy,
)
from mage_ai.data_cleaner.transformer_actions.utils import build_action_variables


class TypeImputeSubRule:
    DATA_SM_UB = 100
    MAX_NULL_SEQ_LENGTH = 4

    def __init__(self, df, column_types, statistics, is_timeseries=False):
        """
        Assumptions of TypeImputeSubRule
        1. df will not contain any empty strings - all empty strings are converted to null types.
        This is handled in ImputeValues.
        2. column_types will contain the correct type value
        3. Every column in df is of dtype object and the entries must be used to infer type.
        This is not always the case, but this assumption simplifies code
        """
        self.df = df
        self.df_columns = df.columns.tolist()
        self.column_types = column_types
        self.is_timeseries = is_timeseries
        self.statistics = statistics

    def accepted_dtypes(self):
        """
        Gets the list of dtypes this subrule accepts and checks
        """
        raise NotImplementedError(
            'Children of TypeImputeSubRule must override \'accepted_dtypes()\''
        )

    def evaluate(self, column):
        """
        Gets the imputation strategy for the given column
        """
        raise NotImplementedError('Children of TypeImputeSubRule must override \'evaluate()\'')


class NumericalImputeSubRule(TypeImputeSubRule):
    ACCEPTED_DTYPES = NUMBER_TYPES
    SKEW_UB = 0.7
    AVG_OR_MED_EMPTY_UB = {'small': 0.3, 'large': 0.5}

    def accepted_dtypes(self):
        return self.ACCEPTED_DTYPES

    def evaluate(self, column):
        """
        Rule:
        1. If there are no null entries, no suggestion
        2. If the dataset was identified as timeseries, suggest sequential imputation
        3. If the number of nonnull entries is sell than DATA_SM_UB, use the
           small dataset bound; else use the large dataset bound
        3a. If the null value rate of the column is greater than AVG_OR_MED_EMPTY_UB
           (which can vary for small vs large dataset), suggest no imputation (not enough values)
        3b. If null value rate is less that AVG_OR_MED_EMPTY_UB and skew is less than SKEW_UB
           suggest imputing with mean value; else impute with median value
        """
        if self.statistics[f'{column}/null_value_rate'] == 0:
            return ImputationStrategy.NOOP
        elif (
            self.is_timeseries
            and self.statistics[f'{column}/max_null_seq'] <= self.MAX_NULL_SEQ_LENGTH
            and self.df[column].notna().iloc[0]
        ):
            return ImputationStrategy.SEQ
        else:
            if self.statistics[f'{column}/count'] <= self.DATA_SM_UB:
                avg_or_med_empty_ub = self.AVG_OR_MED_EMPTY_UB['small']
            else:
                avg_or_med_empty_ub = self.AVG_OR_MED_EMPTY_UB['large']
            if self.statistics[f'{column}/null_value_rate'] <= avg_or_med_empty_ub:
                if abs(self.df[column].skew()) < self.SKEW_UB:
                    return ImputationStrategy.AVERAGE
                else:
                    return ImputationStrategy.MEDIAN
        return ImputationStrategy.CONSTANT


class CategoricalImputeSubRule(TypeImputeSubRule):
    ACCEPTED_DTYPES = CATEGORICAL_TYPES
    RAND_EMPTY_UB = 0.3
    MODE_PROP_LB = 0.4

    def accepted_dtypes(self):
        return self.ACCEPTED_DTYPES

    def evaluate(self, column):
        """
        Rule:
        1. If there are no null entries, no suggestion
        2. If the dataset was identified as timeseries, suggest sequential imputation
        3. Else, if more than MODE_PROP_LB of nonnull entries are a single value, use
           imputation with mode
        4. Else, if less than RAND_EMPTY_UB ratio of entries are null, use random imputation
        5. Else suggest no imputation (no good fit)
        """
        if self.statistics[f'{column}/null_value_rate'] == 0:
            return ImputationStrategy.NOOP
        elif (
            self.is_timeseries
            and self.statistics[f'{column}/max_null_seq'] <= self.MAX_NULL_SEQ_LENGTH
            and self.df[column].notna().iloc[0]
        ):
            return ImputationStrategy.SEQ
        elif self.statistics[f'{column}/mode_ratio'] >= self.MODE_PROP_LB:
            return ImputationStrategy.MODE
        elif self.statistics[f'{column}/null_value_rate'] <= self.RAND_EMPTY_UB:
            return ImputationStrategy.RANDOM
        return ImputationStrategy.CONSTANT


class DateTimeImputeSubRule(TypeImputeSubRule):
    ACCEPTED_DTYPES = frozenset((ColumnType.DATETIME,))

    def accepted_dtypes(self):
        return self.ACCEPTED_DTYPES

    def evaluate(self, column):
        """
        Rule:
        1. If there are no null entries, no suggestion
        2. If the dataset was identified as timeseries, suggest sequential imputation
        3. Else suggest no imputation (no good fit)
        """
        if self.statistics[f'{column}/null_value_rate'] == 0:
            return ImputationStrategy.NOOP
        elif (
            self.is_timeseries
            and self.statistics[f'{column}/max_null_seq'] <= self.MAX_NULL_SEQ_LENGTH
            and self.df[column].notna().iloc[0]
        ):
            return ImputationStrategy.SEQ
        return ImputationStrategy.CONSTANT


class ListImputeSubRule(TypeImputeSubRule):
    ACCEPTED_DTYPES = frozenset((ColumnType.LIST,))
    RAND_EMPTY_UB = 0.3
    MODE_PROP_LB = 0.4

    def accepted_dtypes(self):
        return self.ACCEPTED_DTYPES

    def evaluate(self, column):
        """
        Rule:
        1. If there are no null entries, no suggestion
        2. If the dataset was identified as timeseries, suggest sequential imputation
        3. Else, if more than MODE_PROP_LB of nonnull entries are a single value, use
           imputation with mode
        4. Else, if less than RAND_EMPTY_UB ratio of entries are null, use random imputation
        5. Else suggest no imputation (no good fit)
        """
        if self.statistics[f'{column}/null_value_rate'] == 0:
            return ImputationStrategy.NOOP
        elif (
            self.is_timeseries
            and self.statistics[f'{column}/max_null_seq'] <= self.MAX_NULL_SEQ_LENGTH
            and self.df[column].notna().iloc[0]
        ):
            return ImputationStrategy.SEQ
        elif self.statistics[f'{column}/mode_ratio'] >= self.MODE_PROP_LB:
            return ImputationStrategy.MODE
        elif self.statistics[f'{column}/null_value_rate'] <= self.RAND_EMPTY_UB:
            return ImputationStrategy.RANDOM
        return ImputationStrategy.CONSTANT


class StringImputeSubRule(TypeImputeSubRule):
    # TODO: Give list type its own imputation strategy
    ACCEPTED_DTYPES = STRING_TYPES
    RAND_EMPTY_UB = 0.3
    MODE_PROP_LB = 0.4

    def accepted_dtypes(self):
        return self.ACCEPTED_DTYPES

    def evaluate(self, column):
        """
        Rule:
        1. If there are no null entries, no suggestion
        2. If the dataset was identified as timeseries, suggest sequential imputation
        3. Else, if more than MODE_PROP_LB of nonnull entries are a single value, use
           imputation with mode
        4. Else, if less than RAND_EMPTY_UB ratio of entries are null, use random imputation
        5. Else suggest no imputation (no good fit)
        """
        if self.statistics[f'{column}/null_value_rate'] == 0:
            return ImputationStrategy.NOOP
        if (
            self.is_timeseries
            and self.statistics[f'{column}/max_null_seq'] <= self.MAX_NULL_SEQ_LENGTH
            and self.df[column].notna().iloc[0]
        ):
            return ImputationStrategy.SEQ
        elif self.statistics[f'{column}/mode_ratio'] >= self.MODE_PROP_LB:
            return ImputationStrategy.MODE
        elif self.statistics[f'{column}/null_value_rate'] <= self.RAND_EMPTY_UB:
            return ImputationStrategy.RANDOM
        return ImputationStrategy.CONSTANT


class ImputeValues(BaseRule):
    RULESET = (
        CategoricalImputeSubRule,
        DateTimeImputeSubRule,
        ListImputeSubRule,
        NumericalImputeSubRule,
        StringImputeSubRule,
    )
    ROW_KEPT_LB = 0.7
    TIMESERIES_NULL_RATIO_MAX = 0.1
    STRATEGY_BLACKLIST = [ImputationStrategy.NOOP, ImputationStrategy.ROW_RM]

    def __init__(self, df, column_types, statistics, custom_config={}):
        super().__init__(df, column_types, statistics, custom_config=custom_config)
        self.exact_dtypes = self.get_exact_dtypes()
        self.strategy_cache = {
            ImputationStrategy.AVERAGE: {'entries': []},
            ImputationStrategy.CONSTANT: {'entries': []},
            ImputationStrategy.MEDIAN: {'entries': []},
            ImputationStrategy.MODE: {'entries': []},
            ImputationStrategy.NOOP: {'entries': []},
            ImputationStrategy.RANDOM: {'entries': []},
            ImputationStrategy.ROW_RM: {'entries': []},
            ImputationStrategy.SEQ: {'entries': []},
        }
        if self.statistics['is_timeseries']:
            self.is_timeseries = True
            self.strategy_cache[ImputationStrategy.SEQ]['timeseries_index'] = self.statistics[
                'timeseries_index'
            ]
        else:
            self.is_timeseries = False
        self.hydrate_rules()

    def build_suggestions(self):
        suggestions = []
        if self.strategy_cache[ImputationStrategy.ROW_RM].get('num_missing'):
            strategy_cache_entry = self.strategy_cache[ImputationStrategy.ROW_RM]
            suggestions.append(
                self.construct_suggestion_from_strategy(
                    ImputationStrategy.ROW_RM, strategy_cache_entry
                )
            )
        else:
            for strategy in self.strategy_cache:
                strategy_cache_entry = self.strategy_cache[strategy]
                if (
                    strategy not in self.STRATEGY_BLACKLIST
                    and len(strategy_cache_entry['entries']) != 0
                ):
                    suggestions.append(
                        self.construct_suggestion_from_strategy(strategy, strategy_cache_entry)
                    )
        return suggestions

    def construct_suggestion_from_strategy(self, strategy, strategy_cache_entry):
        title = 'Fill in missing values'
        message = ''
        action_type = None
        action_arguments = []
        action_code = ''
        action_options = {}
        action_variables = {}
        axis = None
        outputs = []

        if strategy == ImputationStrategy.AVERAGE:
            message = 'For each column, fill missing entries with the average value.'
            action_arguments = strategy_cache_entry['entries']
            action_type = ActionType.IMPUTE
            axis = Axis.COLUMN
            action_options = {'strategy': strategy}
            action_variables = build_action_variables(
                self.df, self.column_types, strategy_cache_entry['entries']
            )
        elif strategy == ImputationStrategy.CONSTANT:
            message = 'Fill missing values with a placeholder to mark them as missing.'
            action_arguments = strategy_cache_entry['entries']
            action_type = ActionType.IMPUTE
            axis = Axis.COLUMN
            action_options = {
                'strategy': strategy,
            }
            action_variables = build_action_variables(
                self.df, self.column_types, strategy_cache_entry['entries']
            )
        elif strategy == ImputationStrategy.MEDIAN:
            message = 'For each column, fill missing entries with the median value.'
            action_arguments = strategy_cache_entry['entries']
            action_type = ActionType.IMPUTE
            axis = Axis.COLUMN
            action_options = {'strategy': strategy}
            action_variables = build_action_variables(
                self.df, self.column_types, strategy_cache_entry['entries']
            )
        elif strategy == ImputationStrategy.MODE:
            message = 'For each column, fill missing entries with the most frequent value.'
            action_arguments = strategy_cache_entry['entries']
            action_type = ActionType.IMPUTE
            axis = Axis.COLUMN
            action_options = {'strategy': 'mode'}
            action_variables = build_action_variables(
                self.df, self.column_types, strategy_cache_entry['entries']
            )
        elif strategy == ImputationStrategy.RANDOM:
            message = 'For each column, fill missing entries with randomly sampled values.'
            action_arguments = strategy_cache_entry['entries']
            action_type = ActionType.IMPUTE
            axis = Axis.COLUMN
            action_options = {'strategy': strategy}
            action_variables = build_action_variables(
                self.df, self.column_types, strategy_cache_entry['entries']
            )
        elif strategy == ImputationStrategy.ROW_RM:
            num_missing = strategy_cache_entry['num_missing']
            title = 'Remove rows with missing entries'
            message = f'Delete {num_missing} rows to remove all missing values from the dataset.'
            action_arguments = self.df_columns
            action_type = ActionType.FILTER
            axis = Axis.ROW
            action_variables = build_action_variables(self.df, self.column_types, self.df_columns)
            map_cols = map(wrap_column_name, self.df_columns)
            action_code = ' and '.join(map(lambda name: f'{name} != null', map_cols))
        elif strategy == ImputationStrategy.SEQ:
            message = 'Fill missing entries using the previously occurring entry in the timeseries.'
            action_arguments = strategy_cache_entry['entries']
            action_type = ActionType.IMPUTE
            axis = Axis.COLUMN
            action_options = {
                'strategy': strategy,
                'timeseries_index': strategy_cache_entry['timeseries_index'],
            }
            action_variables = build_action_variables(
                self.df, self.column_types, strategy_cache_entry['entries']
            )

        return self._build_transformer_action_suggestion(
            title,
            message,
            action_type,
            action_arguments,
            action_code,
            action_options,
            action_variables,
            axis,
            outputs,
        )

    def evaluate(self):
        if self.df.empty:
            return []
        non_null_rows = self.df.notna().all(axis=1).sum()
        ratio_rows_kept = non_null_rows / len(self.df)
        if ratio_rows_kept == 1:
            self.strategy_cache[ImputationStrategy.NOOP]['entries'].extend(self.df_columns)
        elif ratio_rows_kept >= self.ROW_KEPT_LB:
            self.strategy_cache[ImputationStrategy.ROW_RM]['num_missing'] = (
                len(self.df) - non_null_rows
            )
        else:
            for column in self.df_columns:
                dtype = self.column_types[column]
                rule = self.rule_map[dtype]
                self.strategy_cache[rule.evaluate(column)]['entries'].append(column)
        return self.build_suggestions()

    def get_exact_dtypes(self):
        def _get_exact_dtype(column):
            dropped = self.df[column].dropna(axis=0)
            try:
                return type(dropped.iloc[0])
            except IndexError:
                return None

        exact_dtypes = {column: _get_exact_dtype(column) for column in self.df_columns}
        return exact_dtypes

    def hydrate_rules(self):
        self.rules = list(
            map(
                lambda x: x(self.df, self.column_types, self.statistics, self.is_timeseries),
                self.RULESET,
            )
        )

        self.rule_map = {}
        for dtype in ColumnType:
            rule_iterator = iter(self.rules)
            curr_rule = next(rule_iterator)
            while dtype not in curr_rule.accepted_dtypes():
                try:
                    curr_rule = next(rule_iterator)
                except StopIteration:
                    raise RuntimeError(f'No rule found to handle imputation of type {dtype}')
            self.rule_map[dtype] = curr_rule



from flask import Flask, render_template, request
from flask_cors import CORS
from mage_ai.data_cleaner.column_types.column_type_detector import infer_column_types
from mage_ai.data_cleaner.data_cleaner import analyze, clean as clean_data
from mage_ai.data_cleaner.pipelines.base import BasePipeline
from mage_ai.data_cleaner.shared.utils import clean_dataframe
from mage_ai.data_cleaner.transformer_actions.utils import generate_action_titles
from mage_ai.server.client.mage import Mage
from mage_ai.server.constants import SERVER_HOST, SERVER_PORT
from mage_ai.server.data.models import FeatureSet, Pipeline
from mage_ai.shared.logger import VerboseFunctionExec
from mage_ai.shared.parsers import encode_complex
import argparse
import json
import logging
import os
import simplejson
import sys
import threading
import traceback


log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)

app = Flask(
    __name__, static_url_path='', static_folder='frontend_dist', template_folder='frontend_dist'
)

CORS(app, resources={r'/*': {'origins': '*'}})

thread = None
api_key = None


class ServerConfig:
    def __init__(self):
        self.server_base_path = None
        self.server_url_params = None


server_config = ServerConfig()


def rescue_errors(endpoint, error_code=500):
    def handler(*args, **kwargs):
        try:
            response = endpoint(*args, **kwargs)
        except Exception as err:
            exception = str(err)
            response_obj = dict(
                error=dict(
                    code=error_code,
                    errors=traceback.format_stack(),
                    exception=exception,
                    message=traceback.format_exc(),
                    type=None,
                ),
            )
            response = app.response_class(
                response=json.dumps(response_obj),
                status=200,
                mimetype='application/json',
            )
            log.exception('An error was caught and reported to the client: ')
        return response

    return handler


@app.route('/', endpoint='index')
@app.route('/datasets', endpoint='dataset_list')
@rescue_errors
def index():
    return render_template('index.html')


@app.route('/datasets/<id>', endpoint='dataset_detail')
@rescue_errors
def dataset(id):
    return render_template('index.html')


@app.route('/process', methods=['POST'], endpoint='process')
@rescue_errors
def process():
    """
    request: {
        id: string (feature set id)
        clean: boolean
    }

    response: {
        id,
        metadata,
        sample_data,
        statistics,
        insights,
        suggestions
    }
    """
    request_data = request.json
    if not request_data:
        request_data = request.form

    id = request_data['id']
    if not id:
        return

    feature_set = FeatureSet(id=id, api_key=api_key)
    df = feature_set.data
    metadata = feature_set.metadata

    if request_data.get('clean', True):
        result = clean_data(df)
    else:
        result = analyze(df)

    feature_set.write_files(result)

    column_types = result['column_types']
    metadata['column_types'] = column_types

    feature_set.metadata = metadata

    response = app.response_class(
        response=simplejson.dumps(feature_set.to_dict(), ignore_nan=True),
        status=200,
        mimetype='application/json',
    )
    return response


@app.route('/feature_sets', endpoint='feature_sets')
@rescue_errors
def feature_sets():
    """
    response: [
        {
            id,
            metadata,
        }
    ]
    """
    feature_sets = list(map(lambda fs: fs.to_dict(detailed=False), FeatureSet.objects()))
    valid_feature_sets = [
        f
        for f in feature_sets
        if set(['column_types', 'statistics']).issubset(f['metadata'].keys())
    ]
    response = app.response_class(
        response=simplejson.dumps(valid_feature_sets, ignore_nan=True),
        status=200,
        mimetype='application/json',
    )
    return response


@app.route('/feature_sets/<id>', endpoint='feature_sets_get')
@rescue_errors
def feature_set(id):
    """
    response: [
        {
            id,
            metadata,
        }
    ]
    """
    if not FeatureSet.is_valid_id(id):
        raise RuntimeError(f'Unknown feature set id: {id}')
    feature_set = FeatureSet(id=id)
    response = app.response_class(
        response=simplejson.dumps(feature_set.to_dict(), ignore_nan=True),
        status=200,
        mimetype='application/json',
    )
    return response


@app.route('/feature_sets/<id>/downloads/', endpoint='feature_set_download', methods=['POST'])
@rescue_errors
def feature_set_download(id):
    if not FeatureSet.is_valid_id(id):
        raise RuntimeError(f'Unknown feature set id: {id}')
    feature_set = FeatureSet(id=id)

    name = feature_set.metadata['name']
    name = name.replace(' ', '_')

    return app.response_class(
        response=feature_set.data.to_csv(index=False),
        status=200,
        mimetype='text/csv',
        headers={
            'Content-Disposition': f'attachment; filename="{name}.csv"',
        },
    )


@app.route('/feature_sets/<id>/versions/<version>', endpoint='feature_set_versions_get')
@rescue_errors
def feature_set_version(id, version):
    """
    response: [
        {
            id,
            metadata,
        }
    ]
    """
    if not FeatureSet.is_valid_id(id):
        raise RuntimeError(f'Unknown feature set id: {id}')
    feature_set = FeatureSet(id=id)
    response = app.response_class(
        response=simplejson.dumps(feature_set.to_dict(version=version), ignore_nan=True),
        status=200,
        mimetype='application/json',
    )
    return response


@app.route('/feature_sets/<id>', methods=['PUT'], endpoint='feature_sets_put')
@rescue_errors
def update_feature_set(id):
    """
    request: {
        metadata,
        statistics,
        insights,
        suggestions,
    }

    response: {
        id,
        metadata,
        pipeline,
        sample_data,
        statistics,
        insights,
        suggestions,
    }
    """
    request_data = request.json
    feature_set = FeatureSet(id=id)
    feature_set.write_files(request_data)
    response = app.response_class(
        response=simplejson.dumps(feature_set.to_dict(), ignore_nan=True),
        status=200,
        mimetype='application/json',
    )
    return response


@app.route('/pipelines', endpoint='pipelines')
@rescue_errors
def pipelines():
    """
    response: [
        {
            id,
            pipeline_actions,
        }
    ]
    """
    pipelines = list(map(lambda p: p.to_dict(detailed=False), Pipeline.objects()))
    response = app.response_class(
        response=simplejson.dumps(
            pipelines,
            default=encode_complex,
            ignore_nan=True,
        ), status=200, mimetype='application/json'
    )
    return response


@app.route('/pipelines/<id>', endpoint='pipelines_get')
@rescue_errors
def pipeline(id):
    """
    response: {
        id,
        actions,
    }
    """
    if not Pipeline.is_valid_id(id):
        raise RuntimeError(f'Unknown pipeline id: {id}')
    pipeline = Pipeline(id=id)
    response = app.response_class(
        response=simplejson.dumps(
            pipeline.to_dict(),
            default=encode_complex,
            ignore_nan=True,
        ),
        status=200,
        mimetype='application/json',
    )
    return response


@app.route('/pipelines/<id>', methods=['PUT'], endpoint='pipelines_put')
@rescue_errors
def update_pipeline(id):
    """
    request: {
        actions,
    }

    response: {
        id,
        actions,
    }
    """
    request_data = request.json
    pipeline = Pipeline(id=id)
    actions = request_data.get('actions', [])
    actions = generate_action_titles(actions)
    clean_pipeline = BasePipeline(actions=actions)
    # 1. Transform the data
    # 2. Recalculate stats and suggestions
    feature_set_id = pipeline.metadata.get('feature_set_id')
    if feature_set_id is not None:
        feature_set = FeatureSet(id=feature_set_id)
        df = feature_set.data_orig
        ctypes = infer_column_types(df, column_types=feature_set.metadata.get('column_types', {}))
        df = clean_dataframe(df, ctypes, False)
        df_transformed = clean_pipeline.transform(df, auto=False)
        result = clean_data(
            df_transformed,
            column_types=ctypes,
            transform=False,
            rule_configs=feature_set.cleaning_rule_configs,
            verbose=True,
        )
        prev_version = len(pipeline.pipeline.actions)
        pipeline.pipeline = clean_pipeline
        feature_set.write_files(result, prev_version=prev_version)
    else:
        pipeline.pipeline = clean_pipeline

    if api_key is not None:
        pipeline.sync_pipeline(api_key)

    response = app.response_class(
        response=simplejson.dumps(
            pipeline.to_dict(),
            default=encode_complex,
            ignore_nan=True,
        ),
        status=200,
        mimetype='application/json',
    )
    return response


@app.route('/status')
@rescue_errors
def get_status():
    return app.response_class(
        response=json.dumps(dict(
            api_key_provided=api_key is not None,
        )),
        status=200,
        mimetype='application/json',
    )


def clean_df(df, name, verbose=False):
    feature_set = FeatureSet(df=df, name=name)

    result = clean_data(df, verbose=verbose)

    feature_set.write_files(result)
    return (feature_set, result['df'])


def clean_df_with_pipeline(
    df, id=None, path=None, remote_id=None, mage_api_key=None, verbose=False
):
    pipeline = None
    if id is not None:
        with VerboseFunctionExec(f'Loading pipeline with uuid \'{id}\' from file', verbose=verbose):
            pipeline = Pipeline(id=id).pipeline
    elif path is not None:
        with VerboseFunctionExec(f'Loading pipeline from path \'{path}\'', verbose=verbose):
            pipeline = Pipeline(path=path).pipeline
    elif remote_id is not None:
        with VerboseFunctionExec(
            f'Loading pipeline \'{remote_id}\' from Mage servers', verbose=verbose
        ):
            final_key = mage_api_key if mage_api_key is not None else api_key
            pipeline = BasePipeline(actions=Mage().get_pipeline_actions(remote_id, final_key))
    if pipeline is None:
        log.error('Please provide a valid pipeline id or config path.')
        return df
    pipeline.verbose = verbose
    return pipeline.transform(df, auto=False)


def connect_df(df, name, verbose=False):
    feature_set = FeatureSet(df=df, name=name)

    result = clean_data(df, transform=False, verbose=verbose)

    feature_set.write_files(result, write_orig_data=True)
    return (feature_set, df)


class ThreadWithTrace(threading.Thread):
    def __init__(self, *args, **keywords):
        threading.Thread.__init__(self, *args, **keywords)
        self.killed = False

    def start(self):
        self.__run_backup = self.run
        self.run = self.__run
        threading.Thread.start(self)

    def __run(self):
        sys.settrace(self.globaltrace)
        self.__run_backup()
        self.run = self.__run_backup

    def globaltrace(self, frame, event, arg):
        if event == 'call':
            return self.localtrace
        else:
            return None

    def localtrace(self, frame, event, arg):
        if self.killed:
            if event == 'line':
                raise SystemExit()
        return self.localtrace

    def kill(self):
        self.killed = True


def launch(mage_api_key=None, host=SERVER_HOST, port=SERVER_PORT) -> None:
    global thread
    sync_pipelines(mage_api_key)

    app_kwargs = {
        'debug': False,
        'host': host,
        'port': port,
    }
    thread = ThreadWithTrace(target=app.run, kwargs=app_kwargs, daemon=True)
    thread.start()
    print(f'Mage running on host and port {host}:{port}')
    return thread


def sync_pipelines(mage_api_key=None):
    global api_key
    if mage_api_key:
        api_key = mage_api_key

    if api_key is not None:
        print('Syncing pipelines with cloud database.', end='')
        local_pipelines = Pipeline.objects()
        for pipeline in local_pipelines:
            print('.', end='')
            pipeline.sync_pipeline(api_key)


def kill():
    if thread is not None:
        thread.kill()
        thread.join()
        try:
            alive = thread.isAlive()
        except AttributeError:
            # thread's method isAlive() was renamed to is_alive() in python version 3.9
            alive = thread.is_alive()
        if not alive:
            print('Flask server is terminated')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, default=None)
    parser.add_argument('--port', type=int, default=None)
    args = parser.parse_args()

    app.debug = os.getenv('DEBUG', True)
    host = args.host or os.getenv('HOST')
    port = args.port or os.getenv('PORT', SERVER_PORT)

    args = dict(port=port)
    if host:
        args['host'] = host

    print(f'Mage running on host and port {host}:{port}')

    app.run(**args)



from croniter import croniter
from datetime import datetime, timedelta
from mage_ai.data_preparation.logger_manager import LoggerManager
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.orchestration.db import Session, session
from mage_ai.shared.array import find
from mage_ai.shared.hash import ignore_keys, index_by
from mage_ai.shared.strings import camel_to_snake_case
from sqlalchemy import Column, DateTime, Enum, ForeignKey, Integer, JSON, String, Table
from sqlalchemy.ext.declarative import declared_attr, declarative_base
from sqlalchemy.orm import relationship, validates
from sqlalchemy.orm.collections import InstrumentedList
from sqlalchemy.sql import func
from typing import Dict, List
import enum

Base = declarative_base()
Base.query = Session.query_property()
Base.select = Session.query


class BaseModel(Base):
    __abstract__ = True

    @declared_attr
    def __tablename__(cls):
        return camel_to_snake_case(cls.__name__)

    id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
    )
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now())

    @classmethod
    def create(self, **kwargs):
        model = self(**kwargs)
        model.save()
        return model

    def save(self, commit=True) -> None:
        session.add(self)
        if commit:
            try:
                session.commit()
            except Exception as e:
                session.rollback()
                raise e

    def update(self, **kwargs) -> None:
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
        session.commit()

    def delete(self, commit: bool = True) -> None:
        session.delete(self)
        if commit:
            session.commit()

    def refresh(self):
        session.refresh(self)

    def to_dict(self, include_attributes=[]) -> Dict:
        def __format_value(value):
            if type(value) is datetime:
                return str(value)
            elif type(value) is InstrumentedList:
                return [__format_value(v) for v in value]
            elif hasattr(value, 'to_dict'):
                return value.to_dict()
            return value
        obj_dict = {c.name: __format_value(getattr(self, c.name)) for c in self.__table__.columns}
        if include_attributes is not None and len(include_attributes) > 0:
            for attr in include_attributes:
                if hasattr(self, attr):
                    obj_dict[attr] = __format_value(getattr(self, attr))
        return obj_dict


pipeline_schedule_event_matcher_association_table = Table(
    'pipeline_schedule_event_matcher_association',
    Base.metadata,
    Column('pipeline_schedule_id', ForeignKey('pipeline_schedule.id')),
    Column('event_matcher_id', ForeignKey('event_matcher.id')),
)


class PipelineSchedule(BaseModel):
    class ScheduleStatus(str, enum.Enum):
        ACTIVE = 'active'
        INACTIVE = 'inactive'

    class ScheduleType(str, enum.Enum):
        TIME = 'time'
        EVENT = 'event'

    class ScheduleInterval(str, enum.Enum):
        ONCE = '@once'
        HOURLY = '@hourly'
        DAILY = '@daily'
        WEEKLY = '@weekly'
        MONTHLY = '@monthly'

    name = Column(String(255))
    pipeline_uuid = Column(String(255))
    schedule_type = Column(Enum(ScheduleType))
    start_time = Column(DateTime(timezone=True), default=None)
    schedule_interval = Column(String(50))
    status = Column(Enum(ScheduleStatus), default=ScheduleStatus.INACTIVE)
    variables = Column(JSON)

    pipeline_runs = relationship('PipelineRun', back_populates='pipeline_schedule')

    event_matchers = relationship(
        'EventMatcher',
        secondary=pipeline_schedule_event_matcher_association_table,
        back_populates='pipeline_schedules'
    )

    @property
    def pipeline_runs_count(self) -> int:
        return len(self.pipeline_runs)

    @validates('schedule_interval')
    def validate_schedule_interval(self, key, schedule_interval):
        if schedule_interval not in [e.value for e in self.__class__.ScheduleInterval]:
            if not croniter.is_valid(schedule_interval):
                raise ValueError('Cron expression is invalid.')

        return schedule_interval

    @classmethod
    def active_schedules(self) -> List['PipelineSchedule']:
        return self.query.filter(self.status == self.ScheduleStatus.ACTIVE).all()

    def current_execution_date(self) -> datetime:
        if self.schedule_interval is None:
            return None

        now = datetime.now()
        if self.schedule_interval == '@once':
            return now
        elif self.schedule_interval == '@daily':
            return now.replace(second=0, microsecond=0, minute=0, hour=0)
        elif self.schedule_interval == '@hourly':
            return now.replace(second=0, microsecond=0, minute=0)
        elif self.schedule_interval == '@weekly':
            return now.replace(second=0, microsecond=0, minute=0, hour=0) - \
                timedelta(days=now.weekday())
        elif self.schedule_interval == '@monthly':
            return now.replace(second=0, microsecond=0, minute=0, hour=0, day=1)
        else:
            cron_itr = croniter(self.schedule_interval, now)
            return cron_itr.get_prev(datetime)

    def should_schedule(self) -> bool:
        if self.status != self.__class__.ScheduleStatus.ACTIVE:
            return False
        if self.start_time is not None and datetime.now() < self.start_time:
            return False

        try:
            Pipeline.get(self.pipeline_uuid)
        except Exception:
            return False

        if self.schedule_interval == '@once':
            if len(self.pipeline_runs) == 0:
                return True
        else:
            """
            TODO: Implement other schedule interval checks
            """
            current_execution_date = self.current_execution_date()
            if current_execution_date is None:
                return False
            if not find(lambda x: x.execution_date == current_execution_date, self.pipeline_runs):
                return True
        return False


class PipelineRun(BaseModel):
    class PipelineRunStatus(str, enum.Enum):
        INITIAL = 'initial'
        RUNNING = 'running'
        COMPLETED = 'completed'
        FAILED = 'failed'
        CANCELLED = 'cancelled'

    pipeline_schedule_id = Column(Integer, ForeignKey('pipeline_schedule.id'))
    pipeline_uuid = Column(String(255))
    execution_date = Column(DateTime(timezone=True))
    status = Column(Enum(PipelineRunStatus), default=PipelineRunStatus.INITIAL)
    completed_at = Column(DateTime(timezone=True))
    variables = Column(JSON)

    pipeline_schedule = relationship(PipelineSchedule, back_populates='pipeline_runs')
    block_runs = relationship('BlockRun', back_populates='pipeline_run')

    def __repr__(self):
        return f'PipelineRun(id={self.id}, pipeline_uuid={self.pipeline_uuid},'\
               f' execution_date={self.execution_date})'

    @property
    def block_runs_count(self) -> int:
        return len(self.block_runs)

    @property
    def execution_partition(self) -> str:
        if self.execution_date is None:
            return str(self.pipeline_schedule_id)
        else:
            return '/'.join([
                        str(self.pipeline_schedule_id),
                        self.execution_date.strftime(format='%Y%m%dT%H%M%S'),
                    ])

    @property
    def log_file(self):
        return File.from_path(LoggerManager.get_log_filepath(
            pipeline_uuid=self.pipeline_uuid,
            partition=self.execution_partition,
        ))

    @property
    def pipeline_schedule_name(self):
        return self.pipeline_schedule.name

    @classmethod
    def active_runs(self) -> List['PipelineRun']:
        return self.query.filter(self.status == self.PipelineRunStatus.RUNNING).all()

    @classmethod
    def create(self, **kwargs) -> 'PipelineRun':
        pipeline_run = super().create(**kwargs)
        pipeline_uuid = kwargs.get('pipeline_uuid')
        if pipeline_uuid is not None:
            pipeline = Pipeline.get(pipeline_uuid)
            blocks = pipeline.get_executable_blocks()
            for b in blocks:
                BlockRun.create(
                    pipeline_run_id=pipeline_run.id,
                    block_uuid=b.uuid,
                )
        return pipeline_run

    def all_blocks_completed(self) -> bool:
        return all(b.status == BlockRun.BlockRunStatus.COMPLETED
                   for b in self.block_runs)


class BlockRun(BaseModel):
    class BlockRunStatus(str, enum.Enum):
        INITIAL = 'initial'
        QUEUED = 'queued'
        RUNNING = 'running'
        COMPLETED = 'completed'
        FAILED = 'failed'
        CANCELLED = 'cancelled'

    pipeline_run_id = Column(Integer, ForeignKey('pipeline_run.id'))
    block_uuid = Column(String(255))
    status = Column(Enum(BlockRunStatus), default=BlockRunStatus.INITIAL)
    started_at = Column(DateTime(timezone=True))
    completed_at = Column(DateTime(timezone=True))

    pipeline_run = relationship(PipelineRun, back_populates='block_runs')

    @property
    def log_file(self):
        return File.from_path(LoggerManager.get_log_filepath(
            pipeline_uuid=self.pipeline_run.pipeline_uuid,
            block_uuid=self.block_uuid,
            partition=self.pipeline_run.execution_partition,
        ))

    @classmethod
    def get(self, pipeline_run_id: int = None, block_uuid: str = None) -> 'BlockRun':
        block_runs = self.query.filter(
            BlockRun.pipeline_run_id == pipeline_run_id,
            BlockRun.block_uuid == block_uuid,
        ).all()
        if len(block_runs) > 0:
            return block_runs[0]
        return None

    def get_outputs(self, sample_count: int = None) -> List[Dict]:
        pipeline = Pipeline.get(self.pipeline_run.pipeline_uuid)
        block = pipeline.get_block(self.block_uuid)
        return block.get_outputs(
            execution_partition=self.pipeline_run.execution_partition,
            sample_count=sample_count,
        )


class EventMatcher(BaseModel):
    class EventType(str, enum.Enum):
        AWS_EVENT = 'aws_event'

    event_type = Column(Enum(EventType), default=EventType.AWS_EVENT)
    name = Column(String(255))
    pattern = Column(JSON)

    pipeline_schedules = relationship(
        'PipelineSchedule',
        secondary=pipeline_schedule_event_matcher_association_table,
        back_populates='event_matchers',
    )

    def __repr__(self):
        return f'EventMatcher(id={self.id}, name={self.name}, pattern={self.pattern})'

    @classmethod
    def active_event_matchers(self) -> List['EventMatcher']:
        return self.query.filter(
            EventMatcher.pipeline_schedules.any(
                PipelineSchedule.status == PipelineSchedule.ScheduleStatus.ACTIVE
            )
        ).all()

    @classmethod
    def upsert_batch(self, event_matchers_payload):
        keys_to_ignore = [
            'created_at',
            'id',
            'updated_at',
        ]

        new_arr = []
        existing_arr = []

        pipeline_schedule_ids = []
        for payload in event_matchers_payload:
            pipeline_schedule_ids += payload.get('pipeline_schedule_ids', [])
            if payload.get('id'):
                existing_arr.append(payload)
            else:
                new_arr.append(payload)

        pipeline_schedules_by_id = index_by(
            lambda x: x.id,
            PipelineSchedule.query.filter(
                PipelineSchedule.id.in_(pipeline_schedule_ids),
            ).all(),
        )

        event_matchers_and_pipeline_schedule_ids = []
        event_matchers_by_id = index_by(
            lambda x: x.id,
            self.query.filter(
                self.id.in_([p['id'] for p in existing_arr]),
            ).all(),
        )
        for payload in existing_arr:
            ids = payload.pop('pipeline_schedule_ids', None)
            event_matcher = event_matchers_by_id[payload['id']]
            event_matcher.update(**ignore_keys(payload, keys_to_ignore))
            event_matchers_and_pipeline_schedule_ids.append((event_matcher, ids))

        for payload in new_arr:
            ids = payload.pop('pipeline_schedule_ids', None)
            event_matcher = self.create(**ignore_keys(payload, keys_to_ignore))
            event_matchers_and_pipeline_schedule_ids.append((event_matcher, ids))

        for event_matcher, ids in event_matchers_and_pipeline_schedule_ids:
            if ids is not None:
                ps = [pipeline_schedules_by_id[i] for i in [int(i) for i in ids]]
                event_matcher.update(pipeline_schedules=ps)

            if event_matcher.event_type == EventMatcher.EventType.AWS_EVENT:
                from mage_ai.services.aws.events.events import update_event_rule_targets
                # For AWS event, update related AWS infra (add trigger to lambda function)
                update_event_rule_targets(event_matcher.name)

        return [t[0] for t in event_matchers_and_pipeline_schedule_ids]

    def active_pipeline_schedules(self) -> List[PipelineSchedule]:
        return [p for p in self.pipeline_schedules
                if p.status == PipelineSchedule.ScheduleStatus.ACTIVE]

    def match(self, config: Dict) -> bool:
        def __match_dict(sub_pattern, sub_config):
            if type(sub_pattern) is not dict or type(sub_config) is not dict:
                return False
            for k in sub_pattern.keys():
                if k not in sub_config:
                    return False
                v = sub_pattern[k]
                if type(v) is list:
                    if sub_config[k] not in v:
                        return False
                elif not __match_dict(v, sub_config[k]):
                    return False
            return True
        return __match_dict(self.pattern, config)



from mage_ai.data_preparation.models.constants import BlockLanguage
from mage_ai.data_preparation.variable_manager import get_variable
from mage_ai.data_preparation.models.block.sql.utils.shared import interpolate_input


def create_upstream_block_tables(loader, block):
    data_provider = block.configuration.get('data_provider')
    database = block.configuration.get('data_provider_database')
    schema = block.configuration.get('data_provider_schema')

    for idx, upstream_block in enumerate(block.upstream_blocks):
        if BlockLanguage.SQL != upstream_block.language or \
           data_provider != upstream_block.configuration.get('data_provider'):

            df = get_variable(
                upstream_block.pipeline.uuid,
                upstream_block.uuid,
                'df',
            )

            loader.export(
                df,
                upstream_block.table_name.upper(),
                database.upper(),
                schema.upper(),
                if_exists='replace',
                verbose=False
            )


def interpolate_input_data(block, query):
    return interpolate_input(
        block,
        query,
        lambda db, schema, tn: f'"{db.upper()}"."{schema.upper()}"."{tn.upper()}"',
    )



from pandas import DataFrame
from typing import Dict, Tuple
import re

"""
Resolves any dependencies and requirements before
a cleaning action is made. All dependency resolution functions return both
- a boolean value describing whether all dependencies are resolved
- a string message describing the error in the case that dependencies aren't resolved
"""


def default_resolution(df: DataFrame, action: Dict) -> Tuple[bool, str]:
    return True, None


def resolve_filter_action(df: DataFrame, action: Dict) -> Tuple[bool, str]:
    for name in df.columns:
        if re.search('\s', name):
            return (
                False,
                'Column name contains whitespace or newline '
                'characters which cannot be used in filter actions',
            )
    return True, None



from mage_ai.data_preparation.models.block.sql import (
    bigquery,
    postgres,
    redshift,
    snowflake,
)
from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.base import DataSource, ExportWritePolicy
from mage_ai.io.config import ConfigFileLoader
from os import path
from time import sleep


def execute_sql_code(block, query):
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = block.configuration.get('data_provider_profile')
    config_file_loader = ConfigFileLoader(config_path, config_profile)

    data_provider = block.configuration.get('data_provider')
    database = block.configuration.get('data_provider_database')
    schema = block.configuration.get('data_provider_schema')
    export_write_policy = block.configuration.get('export_write_policy', ExportWritePolicy.APPEND)

    table_name = block.table_name
    should_query = BlockType.DATA_LOADER == block.type or BlockType.TRANSFORMER == block.type

    if DataSource.BIGQUERY.value == data_provider:
        from mage_ai.io.bigquery import BigQuery

        loader = BigQuery.with_config(config_file_loader)
        bigquery.create_upstream_block_tables(loader, block)

        query_string = bigquery.interpolate_input_data(block, query)
        loader.export(
            None,
            f'{schema}.{table_name}',
            database=database,
            if_exists=export_write_policy,
            query_string=query_string,
            verbose=BlockType.DATA_EXPORTER == block.type,
        )

        if should_query:
            # An error is thrown because the table doesn’t exist until you re-run the query
            # NotFound: 404 Not found: Table database:schema.table_name was not found in location XX
            tries = 0
            while tries < 10:
                sleep(tries)
                tries += 1
                try:
                    result = loader.load(
                        f'SELECT * FROM {database}.{schema}.{table_name}',
                        verbose=False,
                    )
                    return [result]
                except Exception as err:
                    if '404' not in str(err):
                        raise err
    elif DataSource.POSTGRES.value == data_provider:
        from mage_ai.io.postgres import Postgres

        with Postgres.with_config(config_file_loader) as loader:
            postgres.create_upstream_block_tables(loader, block)

            query_string = postgres.interpolate_input_data(block, query)
            loader.export(
                None,
                schema,
                table_name,
                drop_table_on_replace=True,
                if_exists=export_write_policy,
                index=False,
                query_string=query_string,
                verbose=BlockType.DATA_EXPORTER == block.type,
            )

            if should_query:
                return [
                    loader.load(
                        f'SELECT * FROM {schema}.{table_name}',
                        verbose=False,
                    ),
                ]
    elif DataSource.REDSHIFT.value == data_provider:
        from mage_ai.io.redshift import Redshift

        with Redshift.with_config(config_file_loader) as loader:
            redshift.create_upstream_block_tables(loader, block)

            query_string = redshift.interpolate_input_data(block, query)
            loader.export(
                None,
                table_name,
                if_exists=export_write_policy,
                query_string=query_string,
                schema=schema,
                verbose=BlockType.DATA_EXPORTER == block.type,
            )

            if should_query:
                return [
                        loader.load(
                            f'SELECT * FROM {schema}.{table_name}',
                            verbose=False,
                        ),
                    ]
    elif DataSource.SNOWFLAKE.value == data_provider:
        from mage_ai.io.snowflake import Snowflake

        table_name = table_name.upper()
        database = database.upper()
        schema = schema.upper()

        with Snowflake.with_config(config_file_loader, database=database, schema=schema) as loader:
            snowflake.create_upstream_block_tables(loader, block)

            query_string = snowflake.interpolate_input_data(block, query)
            loader.export(
                None,
                table_name,
                database,
                schema,
                if_exists=export_write_policy,
                query_string=query_string,
                verbose=BlockType.DATA_EXPORTER == block.type,
            )

            if should_query:
                return [
                    loader.load(
                        f'SELECT * FROM "{database}"."{schema}"."{table_name}"',
                        verbose=False,
                    ),
                ]

    return []



from enum import Enum
from mage_ai.orchestration.db.database_manager import database_manager
import multiprocessing


def run_scheduler():
    from mage_ai.orchestration.triggers.loop_time_trigger import LoopTimeTrigger

    database_manager.run_migrations()
    LoopTimeTrigger().start()


class SchedulerManager:
    """
    Singleton class to manage scheduler process.
    """

    class SchedulerStatus(str, Enum):
        RUNNING = 'running'
        STOPPED = 'stopped'

    def __init__(self):
        self.scheduler_process: multiprocessing.Process = None

    @property
    def is_alive(self):
        return self.scheduler_process is not None and self.scheduler_process.is_alive()

    def get_status(self):
        if self.is_alive:
            return SchedulerManager.SchedulerStatus.RUNNING
        return SchedulerManager.SchedulerStatus.STOPPED

    def start_scheduler(self):
        if self.is_alive:
            return

        proc = multiprocessing.Process(target=run_scheduler)
        proc.start()
        self.scheduler_process = proc

    def stop_scheduler(self):
        if self.is_alive:
            self.scheduler_process.terminate()
            self.scheduler_process = None


scheduler_manager = SchedulerManager()



from mage_ai.data_preparation.models.pipeline import Pipeline
from typing import Dict
import asyncio


class PipelineExecutor:
    def __init__(self, pipeline: Pipeline):
        self.pipeline = pipeline

    def execute(
        self,
        analyze_outputs: bool = False,
        global_vars: Dict = None,
        run_tests: bool = True,
        update_status: bool = False,
    ) -> None:
        asyncio.run(self.pipeline.execute(
            analyze_outputs=analyze_outputs,
            global_vars=global_vars,
            run_tests=run_tests,
            update_status=update_status,
        ))



from mage_ai.data_preparation.templates.utils import template_env
from mage_ai.services.aws.s3 import s3
import os


class EmrResourceManager:
    def __init__(self, s3_bucket, s3_path_prefix):
        self.s3_bucket = s3_bucket
        self.s3_path_prefix = s3_path_prefix

    @property
    def bootstrap_script_path(self) -> str:
        return os.path.join('s3://', self.s3_bucket, self.bootstrap_script_path_key)

    @property
    def bootstrap_script_path_key(self) -> str:
        return os.path.join(self.s3_path_prefix, 'scripts/emr_bootstrap.sh')

    @property
    def log_uri(self) -> str:
        return os.path.join('s3://', self.s3_bucket, self.s3_path_prefix, 'logs')

    def upload_bootstrap_script(self) -> None:
        bootstrap_script_code = template_env.get_template(
            'pipeline_execution/emr_bootstrap.sh',
        ).render()        
        s3.Client(self.s3_bucket).upload(self.bootstrap_script_path_key, bootstrap_script_code)



from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF


class StringSplit(BaseUDF):
    def execute(self):
        separator = self.options.get('separator')
        part_index = self.options.get('part_index')
        if separator is None or part_index is None:
            raise Exception('Require both `separator` and `part_index` parameters.')
        return self.df[self.arguments[0]].str.split(separator).str[part_index].str.strip()



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.models.constants import LOGS_DIR
import logging
import os


class LoggerManager:
    @classmethod
    def get_logger(
        self,
        repo_path: str = None,
        logs_dir: str = None,
        pipeline_uuid: str = None,
        block_uuid: str = None,
        partition: str = None,
    ):
        log_filepath = self.get_log_filepath(
            repo_path=repo_path,
            logs_dir=logs_dir,
            pipeline_uuid=pipeline_uuid,
            block_uuid=block_uuid,
            partition=partition,
            create_dir=True,
        )

        logger_name_parts = [pipeline_uuid]
        if partition is not None:
            logger_name_parts.append(partition)
        if block_uuid is not None:
            logger_name_parts.append(block_uuid)
        logger_name = '/'.join(logger_name_parts)

        logger = logging.getLogger(logger_name)

        if not len(logger.handlers):
            logger.setLevel(logging.INFO)
            formatter = logging.Formatter(
                '%(asctime)s %(message)s',
                '%Y-%m-%dT%H:%M:%S',
            )

            file_handler = logging.FileHandler(log_filepath)
            file_handler.setLevel(logging.INFO)
            file_handler.setFormatter(formatter)

            logger.addHandler(file_handler)
        return logger

    @classmethod
    def get_log_filepath(
        self,
        repo_path: str = None,
        logs_dir: str = None,
        pipeline_uuid: str = None,
        block_uuid: str = None,
        partition: str = None,
        create_dir: bool = False,
    ):
        repo_path = repo_path or get_repo_path()
        logs_dir = logs_dir or repo_path

        if pipeline_uuid is None:
            raise Exception('Please specify a pipeline uuid in your logger.')

        logs_dir_path = os.path.join(
            logs_dir,
            'pipelines',
            pipeline_uuid,
            LOGS_DIR,
            partition or '',
        )

        if create_dir and not os.path.exists(logs_dir_path):
            os.makedirs(logs_dir_path)

        if block_uuid is None:
            log_filepath = os.path.join(logs_dir_path, 'pipeline.log')
        else:
            log_filepath = os.path.join(logs_dir_path, f'{block_uuid}.log')
        return log_filepath


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """
    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())



from .base import BaseDetailHandler, BaseHandler
from mage_ai.orchestration.db.models import EventMatcher, PipelineSchedule
from mage_ai.orchestration.triggers.event_trigger import EventTrigger
import json


class ApiAwsEventRuleListHandler(BaseHandler):
    def get(self, provider):
        event_rules = []

        if 'aws' == provider:
            from mage_ai.services.aws.events.events import get_all_event_rules
            event_rules = get_all_event_rules()

        self.write(dict(event_rules=event_rules))


class ApiEventHandler(BaseHandler):
    def post(self):
        payload = self.get_payload()
        EventTrigger().run(payload)
        self.write(dict(event=payload))


class ApiEventMatcherDetailHandler(BaseDetailHandler):
    model_class = EventMatcher

    def get(self, event_matcher_id):
        include_attributes = []
        if self.get_bool_argument('include_pipeline_schedules', False):
            include_attributes.append('pipeline_schedules')
        super().get(event_matcher_id, include_attributes=include_attributes)

    def put(self, event_matcher_id):
        event_matcher = self.model_class.query.get(int(event_matcher_id))
        payload = self.get_payload()
        pipeline_schedule_ids = payload.pop('pipeline_schedule_ids', None)
        if pipeline_schedule_ids is not None:
            pipeline_schedules = PipelineSchedule.query.filter(
                PipelineSchedule.id.in_(pipeline_schedule_ids),
            ).all()
            payload['pipeline_schedules'] = pipeline_schedules
        event_matcher.update(**payload)
        self.write_model(event_matcher)

    def delete(self, event_matcher_id):
        super().delete(event_matcher_id)


class ApiEventMatcherListHandler(BaseHandler):
    model_class = EventMatcher

    def get(self):
        event_matchers = EventMatcher.query.all()
        collection = [e.to_dict() for e in event_matchers]
        self.write(dict(event_matchers=collection))
        self.finish()

    def post(self):
        arr = json.loads(self.request.body).get('event_matchers', [])
        if len(arr) == 0:
            arr.append(self.get_payload())

        event_matchers = EventMatcher.upsert_batch(arr)

        if len(event_matchers) >= 2:
            self.write(dict(event_matchers=[em.to_dict() for em in event_matchers]))
        else:
            self.write(dict(event_matcher=event_matchers[0].to_dict()))



from mage_ai.orchestration.db.models import BlockRun, PipelineRun, PipelineSchedule
from mage_ai.server.api.base import BaseHandler
from sqlalchemy.orm import aliased


class ApiPipelineLogListHandler(BaseHandler):
    def get(self, pipeline_uuid):
        pipeline_schedule_ids = self.get_argument('pipeline_schedule_id[]', None)
        if pipeline_schedule_ids:
            pipeline_schedule_ids = pipeline_schedule_ids.split(',')
        else:
            pipeline_schedule_ids = []

        block_uuids = self.get_argument('block_uuid[]', None)
        if block_uuids:
            block_uuids = block_uuids.split(',')
        else:
            block_uuids = []

        pipeline_run_ids = self.get_argument('pipeline_run_id[]', None)
        if pipeline_run_ids:
            pipeline_run_ids = pipeline_run_ids.split(',')
        else:
            pipeline_run_ids = []

        block_run_ids = self.get_argument('block_run_id[]', None)
        if block_run_ids:
            block_run_ids = block_run_ids.split(',')
        else:
            block_run_ids = []

        a = aliased(PipelineRun, name='a')
        b = aliased(PipelineSchedule, name='b')

        columns = [
            a.execution_date,
            a.pipeline_schedule_id,
            a.pipeline_schedule_id,
            a.pipeline_uuid,
        ]

        query = (
            PipelineRun.
            select(*columns).
            join(b, a.pipeline_schedule_id == b.id).
            filter(b.pipeline_uuid == pipeline_uuid)
        )

        if len(pipeline_schedule_ids):
            query = (
                query.
                filter(a.pipeline_schedule_id.in_(pipeline_schedule_ids))
            )

        if len(pipeline_run_ids):
            query = (
                query.
                filter(a.id.in_(pipeline_run_ids))
            )

        pipeline_run_logs = []
        if not len(block_uuids) and not len(block_run_ids):
            rows = query.all()
            for row in rows:
                model = PipelineRun()
                model.execution_date = row.execution_date
                model.pipeline_schedule_id = row.pipeline_schedule_id
                model.pipeline_schedule_id = row.pipeline_schedule_id
                model.pipeline_uuid = row.pipeline_uuid

                pipeline_run_logs.append(model.log_file.to_dict(include_content=True))

        c = aliased(BlockRun, name='c')
        query = (
            BlockRun.
            select(*(columns + [
                c.block_uuid,
            ])).
            join(a, a.id == c.pipeline_run_id).
            join(b, a.pipeline_schedule_id == b.id).
            filter(b.pipeline_uuid == pipeline_uuid)
        )

        if len(block_uuids):
            query = (
                query.
                filter(c.block_uuid.in_(block_uuids))
            )

        if len(block_run_ids):
            query = (
                query.
                filter(c.id.in_(block_run_ids))
            )

        if len(pipeline_schedule_ids):
            query = (
                query.
                filter(a.pipeline_schedule_id.in_(pipeline_schedule_ids))
            )

        if len(pipeline_run_ids):
            query = (
                query.
                filter(a.id.in_(pipeline_run_ids))
            )

        rows = query.all()
        block_run_logs = []
        for row in rows:
            model = PipelineRun()
            model.execution_date = row.execution_date
            model.pipeline_schedule_id = row.pipeline_schedule_id
            model.pipeline_schedule_id = row.pipeline_schedule_id
            model.pipeline_uuid = row.pipeline_uuid

            model2 = BlockRun()
            model2.block_uuid = row.block_uuid
            model2.pipeline_run = model

            block_run_logs.append(model2.log_file.to_dict(include_content=True))

        self.write(dict(logs=[
            dict(
                block_run_logs=block_run_logs,
                pipeline_run_logs=pipeline_run_logs,
            ),
        ]))
